# © Copyright Databand.ai, an IBM Company 2022

from dbnd._vendor import click


def created_deprecated_command(cmd):
    @click.command(
        name=cmd,
        context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
    )
    def f():
        print(f"'dbnd {cmd}' not supported any more, please use 'dbnd-web {cmd}'")

    return f


def add_deprecated_commands(cli):
    for cmd in ["db", "webserver", "command"]:
        cli.add_command(created_deprecated_command(cmd))



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys
import traceback

from functools import wraps

import dbnd

from dbnd._core.utils.timezone import utcnow


logger = logging.getLogger(__name__)


def _format_exception(e_type, e_value, e_traceback):
    return "".join(traceback.format_exception(e_type, value=e_value, tb=e_traceback))


def log_exception_to_server(exception=None):
    try:
        from dbnd._core.current import get_databand_context

        client = get_databand_context().databand_api_client
        if client is None or not client.is_configured():
            return

        e_type, e_value, e_traceback = sys.exc_info()
        if exception:
            e_type, e_value, e_traceback = (
                type(exception),
                exception,
                exception.__traceback__,
            )

        trace = _format_exception(e_type, e_value, e_traceback)

        data = {
            "dbnd_version": dbnd.__version__,
            "source": "tracking-sdk",
            "stack_trace": trace,
            "timestamp": utcnow().isoformat(),
        }
        return client.api_request(endpoint="log_exception", method="POST", data=data)
    except Exception:  # noqa
        logger.debug("Error sending monitoring exception message", exc_info=True)


def capture_tracking_exception(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            log_exception_to_server(e)
            raise e

    return wrapper



# © Copyright Databand.ai, an IBM Company 2022

import sys
import traceback

import attr

from dbnd._core.errors import DatabandError, get_help_msg
from dbnd._core.errors.errors_utils import (
    frame_info_to_str,
    nested_exceptions_str,
    show_exc_info,
)
from dbnd._core.tracking.schemas.tracking_info_objects import ErrorInfo
from dbnd._core.utils.timezone import utcnow
from dbnd.utils.anonymization import secrets_anonymizer


def task_call_source_to_str(task_call_source):
    result = []
    for frame_info in reversed(task_call_source):
        result.append(frame_info_to_str(frame_info))

    return "".join(result).rstrip()


@attr.s
class TaskRunError(object):
    exc_info = attr.ib()  # type: Tuple
    task_run = attr.ib()  # type: TaskRun
    traceback = attr.ib()  # type: str
    exc_time = attr.ib(default=utcnow())  # type: Datetime
    msg = attr.ib(default=None)  # type: str
    airflow_context = attr.ib(default=None)  # type: Dict

    @property
    def exception(self):
        # type:()-> Exception
        return self.exc_info[1]

    @classmethod
    def build_from_ex(self, ex, task_run, exc_info=None):
        exc_info = exc_info or sys.exc_info()
        return TaskRunError(
            exc_info=exc_info,
            traceback=traceback.format_exc(),
            task_run=task_run,
            airflow_context=task_run.airflow_context,
        )

    @classmethod
    def build_from_message(cls, task_run, msg, help_msg, ex_class=DatabandError):
        """
        Builds TaskRunError from string
        TODO: very ugly hack, we need to support TaskRunError without exc_info
        :param task_run:
        :param msg:
        :param help_msg:
        :param ex_class:
        :return: TaskRunError
        """
        try:
            raise ex_class(msg, show_exc_info=False, help_msg=help_msg)
        except DatabandError:
            return TaskRunError(
                exc_info=sys.exc_info(),
                traceback=traceback.format_exc(),
                task_run=task_run,
                airflow_context=task_run.airflow_context,
            )

    def as_error_info(self):
        ex = self.exception
        task = self.task_run.task
        isolate = not task.task_definition.full_task_family.startswith("databand.")
        return ErrorInfo(
            msg=str(ex),
            help_msg=get_help_msg(ex),
            exc_type=self.exc_info[0],
            databand_error=isinstance(ex, DatabandError),
            nested=nested_exceptions_str(ex, limit=1),
            traceback=self.traceback,
            user_code_traceback=secrets_anonymizer.anonymize(
                task.settings.log.format_exception_as_str(
                    exc_info=self.exc_info, isolate=isolate
                )
            ),
            show_exc_info=bool(show_exc_info(ex)),
        )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import TaskEssence
from dbnd._core.errors.base import ConfigLookupError
from dbnd._core.task.task_with_params import _TaskWithParams
from dbnd._core.utils.basics.singleton_context import SingletonContext


class Config(_TaskWithParams, SingletonContext):
    """
    Class for configuration, See ``ConfigClasses``.

    The nice thing is that we can instantiate this class
    and get an object with all the environment variables set.
    This is arguably a bit of a hack.
    """

    task_essence = TaskEssence.CONFIG

    # we need this empty constructor, to support "any" parameters at the ctor auto-completion
    def __init__(self, **kwargs):
        super(Config, self).__init__(**kwargs)

    def __str__(self):
        return self.task_name

    @classmethod
    def from_databand_context(cls, name=None):
        """Syntax sugar for accessing the current config instance."""
        from dbnd._core.current import get_databand_context

        if not name:
            if cls._conf__task_family:
                # using the current cls section name to get the current instance of a class
                name = cls._conf__task_family
            else:
                raise ConfigLookupError(
                    "name is required for retrieving a config instance"
                )

        return get_databand_context().settings.get_config(name)


Config.task_definition.hidden = True



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from dbnd import task
from dbnd.tasks.doctor.doctor_report_builder import DoctorStatusReportBuilder


logger = logging.getLogger(__name__)


@task
def airflow_status():
    """
    Raises error if required tracking plugin is not installed.
    Logs plugins versions if installed.
    """
    report = DoctorStatusReportBuilder("Airflow Status")
    logger.info("Checking Databand+Airflow installation")

    report.log("env.AIRFLOW_HOME", os.environ.get("AIRFLOW_HOME"))
    try:
        import dbnd_airflow

        report.log("dbnd_airflow.version", dbnd_airflow.__version__)
    except ImportError:
        report.log("dbnd_airflows.version", 0)
        logger.error("dbnd_airflow is not installed")

    try:
        import dbnd_airflow_auto_tracking

        report.log(
            "dbnd_airflow_auto_tracking.version", dbnd_airflow_auto_tracking.__version__
        )
    except ImportError:

        report.log("dbnd_airflow_auto_tracking.version", 0)
        logger.error("dbnd_airflow_auto_tracking is not installed")

    try:
        import airflow

        report.log("airflow.version", airflow.__version__)
    except ImportError:
        report.log("airflow.version", 0)
        logger.error("airflow is not installed")

    return report.get_status_str()



# coding: utf-8
# Copyright 2016 Étienne Bersac
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import unittest

import six

from dbnd._vendor.tenacity import _asyncio as tasyncio
from dbnd._vendor.tenacity import retry
from dbnd._vendor.tenacity.tests.test_tenacity import NoIOErrorAfterCount


def asynctest(callable_):
    callable_ = asyncio.coroutine(callable_)

    @six.wraps(callable_)
    def wrapper(*a, **kw):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(callable_(*a, **kw))

    return wrapper


@retry
@asyncio.coroutine
def _retryable_coroutine(thing):
    yield from asyncio.sleep(0.00001)
    thing.go()


class TestAsync(unittest.TestCase):
    @asynctest
    def test_retry(self):
        assert asyncio.iscoroutinefunction(_retryable_coroutine)
        thing = NoIOErrorAfterCount(5)
        yield from _retryable_coroutine(thing)
        assert thing.counter == thing.count

    def test_repr(self):
        repr(tasyncio.AsyncRetrying())


if __name__ == '__main__':
    unittest.main()



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import typing

from typing import Any, Dict, List, Optional, Type
from uuid import UUID

import more_itertools
import six

from dbnd._core.constants import RESULT_PARAM, TaskEssence, _TaskParamContainer
from dbnd._core.current import (
    get_databand_context,
    get_databand_run,
    try_get_current_task,
)
from dbnd._core.parameter import build_user_parameter_value
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import (
    ParameterFilters,
    Parameters,
    ParameterValue,
)
from dbnd._core.task.base_task import _BaseTask
from dbnd._core.task.task_mixin import _TaskCtrlMixin
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_build.task_signature import Signature, user_friendly_signature
from dbnd._core.task_build.task_source_code import TaskSourceCode
from dbnd._core.task_ctrl.task_ctrl import TrackingTaskCtrl
from dbnd._core.task_ctrl.task_output_builder import windows_drive_re
from dbnd._core.task_ctrl.task_relations import traverse_and_set_target
from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.basics.nothing import NOTHING, is_not_defined
from dbnd._core.utils.callable_spec import CallableSpec, args_to_kwargs
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.uid_utils import get_uuid
from targets import InMemoryTarget, target
from targets.base_target import TargetSource
from targets.target_config import folder
from targets.utils.path import no_trailing_slash


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


def _generate_unique_tracking_signature():
    return Signature(
        "tracking", user_friendly_signature(str(get_uuid())), "unique tracking call"
    )


class TrackingTask(_BaseTask, _TaskCtrlMixin, _TaskParamContainer):
    """
    Task for tracking only use-cases.

    this task should not contain any class body Parameters definition
    """

    task_essence = TaskEssence.TRACKING
    #############################
    # BACKWARD COMPATIBLE TO `Task`
    task_class_version = ""
    task_is_dynamic = False
    task_is_system = False

    @classmethod
    def for_func(cls, task_definition, task_args, task_kwargs, task_name=None):
        # type: (Type[TrackingTask], TaskDefinition, List[Any], Dict[str,Any], str) -> TrackingTask
        """
        Creating the task for a decorated function.
        We use pre-created definition for the function definition.

        Here we create the tasks params from runtime input
        """
        param_values = build_func_parameter_values(
            task_definition, task_args, task_kwargs
        )
        # we need to add RESULT param
        if RESULT_PARAM in task_definition.task_param_defs:
            param = task_definition.task_param_defs[RESULT_PARAM]
            if isinstance(param, FuncResultParameter):
                for param_name in param.names:
                    # we want to get the parameter evolved with the task_definition as owner
                    inner_param = task_definition.task_param_defs[param_name]
                    result_param_value = build_result_param(
                        task_definition.task_passport, param_def=inner_param
                    )

                    param_values.append(result_param_value)

            result_param_value = build_result_param(
                task_definition.task_passport, param_def=param
            )
            param_values.append(result_param_value)

        task_params = Parameters(source="tracking_task", param_values=param_values)

        return cls(
            task_name=task_name or task_definition.task_family,
            task_definition=task_definition,
            task_params=task_params,
        )

    @classmethod
    def for_user_params(
        cls,
        task_definition_uid,  # type: UUID
        task_name,  # type: str
        user_params,  # type: Dict[str,Any]
        task_passport,  # type: TaskPassport
        source_code=None,  # type: TaskSourceCode
        result=True,  # type: Optional[bool]
    ):
        # type: (...) -> TrackingTask

        """
        Creating a customize task from the required params.
        Here we build the params from runtime values and use them to build the task definition.
        """
        param_values = []
        for key, value in six.iteritems(user_params):
            p = build_user_parameter_value(
                name=key, value=value, source=task_passport.full_task_family_short
            )
            param_values.append(p)

        if result:
            result_param_value = build_result_param(task_passport)
            param_values.append(result_param_value)

        task_params = Parameters(
            source=task_passport.full_task_family_short, param_values=param_values
        )

        task_definition = TaskDefinition(
            task_passport=task_passport,
            source_code=source_code,
            external_parameters=task_params,
            task_definition_uid=task_definition_uid,
        )

        return cls(
            task_name=task_name,
            task_definition=task_definition,
            task_params=task_params,
        )

    def __init__(
        self,
        task_name,
        task_definition,
        task_params,
        task_signature_obj=None,
        task_version=None,
    ):
        task_signature_obj = task_signature_obj or _generate_unique_tracking_signature()

        super(TrackingTask, self).__init__(
            task_name=task_name,
            task_definition=task_definition,
            task_signature_obj=task_signature_obj,
            task_params=task_params,
        )

        self.task_definition = task_definition  # type: TaskDefinition
        # we don't have signature for outputs
        self.task_outputs_signature_obj = self.task_signature_obj
        self.ctrl = TrackingTaskCtrl(self)

        self.task_call_source = [
            self.dbnd_context.user_code_detector.find_user_side_frame(1)
        ]
        parent_task = try_get_current_task()
        if parent_task:
            parent_task.descendants.add_child(self.task_id)
            self.task_call_source.extend(parent_task.task_call_source)

            # inherit from parent if it has it
            self.task_version = task_version or parent_task.task_version
            self.task_target_date = parent_task.task_target_date
            self.task_env = parent_task.task_env
            # pass-through parent children scope params
            # task_children_scope_params will be used in case of any Task inside TrackedTask
            # for example tracked task creates Config objects
            self.task_children_scope_params = parent_task.task_children_scope_params
        else:
            # we need better definition of "what we use for tracking"
            self.task_version = task_version or utcnow().strftime("%Y%m%d_%H%M%S")
            self.task_target_date = utcnow().date()
            self.task_env = get_databand_context().env
            self.task_children_scope_params = {}

        self.task_outputs = dict()
        for parameter, value in self._params.get_params_with_value(
            ParameterFilters.OUTPUTS
        ):
            if is_not_defined(value):
                value_as_target = self.build_tracking_output(parameter)
                task_params.update_param_value(parameter.name, value_as_target)

            if isinstance(parameter, FuncResultParameter):
                continue

            # This is used to keep backward compatibility for tracking luigi behaviour
            # This is not something we want to keep, at least not in this form
            value = traverse_and_set_target(value, parameter._target_source(self))
            self.task_outputs[parameter.name] = value

        self.ctrl._initialize_task()

        # so we can be found via task_id
        self.dbnd_context.task_instance_cache.register_task_instance(self)

    def build_tracking_output(self, p):
        return InMemoryTarget(
            path="memory://{value_type}:{task}.{p_name}".format(
                value_type=p.value_type, task=self.task_id, p_name=p.name
            ),
            source=TargetSource(task_id=self.task_id, parameter_name=p.name),
        )

    def get_task_family(self):
        return self.task_definition.task_family

    @property
    def tracker(self):
        return self.current_task_run.tracker

    @property
    def task_dag(self):
        return self.ctrl.task_dag

    @property
    def descendants(self):
        return self.ctrl.descendants

    @property
    def current_task_run(self):
        # type: ()->TaskRun
        return get_databand_run().get_task_run(self.task_id)

    @property
    def task_in_memory_outputs(self):
        return True

    @property
    @cached()
    def _meta_output(self):
        # in some sense this is a duplication of dbnd._core.task_ctrl.task_output_builder.calculate_path
        # but it also breaking an awful abstraction and a lot of inner functions which is good
        task_env = self.task_env
        sep = "/"
        root = no_trailing_slash(str(task_env.root))
        if windows_drive_re.match(root):
            sep = os.sep
        path = sep.join(
            (
                root,
                task_env.env_label,
                str(self.task_target_date),
                self.task_name,
                self.task_name + "_" + self.task_signature,
                "_meta_output",
                "meta",
            )
        )

        # meta_output is directory
        path += sep

        return target(path, folder)

    def _complete(self):
        return None


def build_result_param(task_passport, param_def=None, name=RESULT_PARAM):
    # type: (TaskPassport, Optional[ParameterDefinition], str) -> ParameterValue
    """
    Build results parameter for the task definition, if parameter definition is not specify it will build a naive one.
    """

    if not param_def:
        from targets.values import ObjectValueType

        # naive creation of result param definition - default named "result" and single value
        param_def = parameter.modify(
            name=name, value_type=ObjectValueType
        ).output.build_parameter("inline")

    return ParameterValue(
        parameter=param_def,
        source=task_passport.full_task_family_short,
        source_value=None,
        value=NOTHING,
        parsed=False,
    )


def build_func_parameter_values(task_definition, task_args, task_kwargs):
    # type: (TaskDefinition, List[Any], Dict[str, Any]) -> List[ParameterValue]
    """
    In tracking task we need to build params without definitions.
    Those params value need no calculations and therefore are very easy to construct
    """
    callable_spec = (
        task_definition.task_decorator.get_callable_spec()
    )  # type: CallableSpec
    values = []

    # convert any arg to kwarg if possible
    args, task_kwargs = args_to_kwargs(callable_spec.args, task_args, task_kwargs)

    # the parameter of the * argument
    if callable_spec.varargs:
        # build the parameter value for the varargs
        vargs_param = build_user_parameter_value(
            callable_spec.varargs,
            tuple(args),
            source=task_definition.full_task_family_short,
        )
        values.append(vargs_param)

    # distinguish between the parameter we expect to create vs those we don't
    unknown_kwargs, known_kwargs = more_itertools.partition(
        lambda kwarg: kwarg[0] in callable_spec.args, six.iteritems(task_kwargs)
    )

    # the parameter of the ** argument
    if callable_spec.varkw:
        # build the parameter value for the varkw
        varkw_param = build_user_parameter_value(
            callable_spec.varkw,
            dict(unknown_kwargs),
            source=task_definition.full_task_family_short,
        )
        values.append(varkw_param)

    # Exclude the self param just like it's excluded in DecoratedCallableParamBuilder
    excluded_params = {"self"}

    for name, value in known_kwargs:
        if name in excluded_params:
            continue

        # build the parameters for the expected parameters
        param_value = build_user_parameter_value(
            name, value, source=task_definition.full_task_family_short
        )
        values.append(param_value)

    return values



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._vendor.termcolor import colored


LINE_BRAKE = "\r\n"
MERGE_MSG = "\r\n...\r\n\r\nThe log body is truncated by databand, fetched {head_size} bytes for the `head` and {tail_size} bytes for the `tail` from the whole {total_log_size} bytes of the logs.\r\nControl the log preview length with LoggingConfig.preview_head_bytes and LoggingConfig.preview_tail_bytes \r\n\r\n...\r\n"
EMPTY_LOG_MSG = "Log is empty"
EMPTY_TRUNCATED_LOG_MSG = "Log truncated to empty"
MERGE_MSG_COLOR = "yellow"


class BufferedLogManager:
    """Buffered log body builder, for in-memory log buffering.

    This manager is responsible to handle how a final log body is supposed
    to look like, when there is a limitation to how much of the start and/or of the end
    of the log is needed to be shown.
    """

    terminator = "\n"

    def __init__(self, max_head_bytes=None, max_tail_bytes=None):
        from dbnd._core.settings import LoggingConfig

        self.initial_buffer = []
        self.initial_buffer_size = 0

        self.max_head_bytes = max_head_bytes
        self.max_tail_bytes = max_tail_bytes
        if self.max_head_bytes is None or self.max_tail_bytes is None:
            logger_config = LoggingConfig()
            self.max_head_bytes = max_head_bytes or logger_config.preview_head_bytes
            self.max_tail_bytes = max_tail_bytes or logger_config.preview_tail_bytes

        self.head_buffer = []
        self.current_head_bytes = 0
        self.trimmed_tail_buffer = []
        self.current_tail_bytes = 0

        self.tail_buffer = []

        self.total_log_size = 0

        self.tail_buffer_starts_with_partial_message = False

    @property
    def _is_initial_buffer_not_full(self) -> bool:
        return self.initial_buffer_size <= self.max_head_bytes + self.max_tail_bytes

    def _does_message_fit_in_initial_buffer(self, message_size) -> bool:
        return (
            self.initial_buffer_size + message_size
            <= self.max_head_bytes + self.max_tail_bytes
        )

    def add_log_msg(self, msg: str):
        """Add the message msg to the log body.

        Args:
            msg: current log message we are adding to the log body.
        """
        if self.max_head_bytes <= 0 and self.max_tail_bytes <= 0:
            return

        message_size = len(msg)
        self.total_log_size += message_size

        if self._is_initial_buffer_not_full:
            if self._does_message_fit_in_initial_buffer(message_size):
                self.initial_buffer.append(msg)
                self.initial_buffer_size += message_size
            else:
                self.initial_buffer_size = self.max_head_bytes + self.max_tail_bytes + 1
                for message in self.initial_buffer:
                    self._add_message_to_truncated_buffers(message)

                self._add_message_to_truncated_buffers(msg)
        else:
            self._add_message_to_truncated_buffers(msg)

    @property
    def minimum_acceptable_initial_partial_message_size(self) -> float:
        """The minimum acceptable size of an initial message, in the tail buffer.

        If the initial message's size in the tail buffer is smaller than this size, it gets deleted when requesting
        the log body.
        If it is larger than this size, than the partial message is kept, because of its substantial part out of the
        tail.
        """
        return self.max_tail_bytes * 0.5

    def _add_message_to_trimmed_tail(self, msg: str):
        message_size = len(msg)
        if self.current_tail_bytes + message_size <= self.max_tail_bytes:
            self.current_tail_bytes += message_size
            self.trimmed_tail_buffer.append(msg)
        elif self.max_tail_bytes > 0:
            # We have overflow in the tail
            bytes_to_rotate = message_size
            if self.current_tail_bytes < self.max_tail_bytes:
                bytes_to_rotate = (
                    self.current_tail_bytes + message_size - self.max_tail_bytes
                )
                self.current_tail_bytes = self.max_tail_bytes

            self.trimmed_tail_buffer.append(msg)
            # Rotation of bytes is needed to keep the newest max_tail_bytes bytes in the trimmed buffer
            while bytes_to_rotate > 0:
                if len(self.trimmed_tail_buffer[0]) <= bytes_to_rotate:
                    bytes_to_rotate -= len(self.trimmed_tail_buffer[0])
                    self.trimmed_tail_buffer = self.trimmed_tail_buffer[1:]
                    self.tail_buffer_starts_with_partial_message = False
                else:
                    self.trimmed_tail_buffer[0] = self.trimmed_tail_buffer[0][
                        bytes_to_rotate:
                    ]
                    self.tail_buffer_starts_with_partial_message = True
                    bytes_to_rotate = 0

    def get_log_body(self) -> str:
        """Builds the final log body.

        Returns:
            final log body, with a truncation message, if occurred.
        """
        if (
            self.current_head_bytes == 0
            and len(self.tail_buffer) == 0
            and self.initial_buffer_size > 0
        ):
            parts = [self.initial_buffer]
        else:
            if self.max_tail_bytes > 0:
                self._build_tail_for_preview()

            if (
                self.tail_buffer_starts_with_partial_message
                and len(self.trimmed_tail_buffer[0])
                < self.minimum_acceptable_initial_partial_message_size
            ):
                self.trimmed_tail_buffer.pop(0)

            parts = [self.head_buffer, self.trimmed_tail_buffer]
        # check if all the parts contains information
        if not any(parts):
            if self.max_head_bytes > 0 or self.max_tail_bytes > 0:
                return EMPTY_LOG_MSG
            # all the parts are truncated
            return EMPTY_TRUNCATED_LOG_MSG

        if not all(parts) and self.max_tail_bytes > 0 and self.max_head_bytes > 0:
            # tail buffer is an empty list
            parts = parts[0:1]

        # join each part with new lines
        joined_parts = [LINE_BRAKE.join(part) for part in parts]

        # building the merge message between head and tail
        merge_msg = colored(
            MERGE_MSG.format(
                head_size=self.max_head_bytes,
                tail_size=self.max_tail_bytes,
                total_log_size=self.total_log_size,
            ),
            MERGE_MSG_COLOR,
        )

        return merge_msg.join(joined_parts)

    def _build_tail_for_preview(self):
        self.current_tail_bytes = 0
        self.trimmed_tail_buffer = []
        for message in self.tail_buffer:
            self._add_message_to_trimmed_tail(message)

    def _add_message_to_truncated_buffers(self, msg: str):
        message_size = len(msg)
        if self.current_head_bytes < self.max_head_bytes:
            if self.current_head_bytes + message_size <= self.max_head_bytes:
                self.current_head_bytes += message_size
                self.head_buffer.append(msg)
            elif self.max_head_bytes > 0:
                # handle when there is some space in the head buffer, but we need to cut it off
                head_buffer_size_left = self.max_head_bytes - self.current_head_bytes
                self.head_buffer.append(msg[0:head_buffer_size_left])
                self.current_head_bytes = self.max_head_bytes
                if self.max_tail_bytes > 0:
                    self.tail_buffer_starts_with_partial_message = True
                    self._add_message_to_tail(msg[head_buffer_size_left:])

        elif self.max_tail_bytes > 0:
            self._add_message_to_tail(msg)

    def _add_message_to_tail(self, msg: str):
        self.tail_buffer.append(msg)



# © Copyright Databand.ai, an IBM Company 2022

_SHOWN_ATTR = "_dbnd_shown"


def set_shown(ex):
    setattr(ex, _SHOWN_ATTR, True)


def is_shown(ex):
    if not ex:
        return False
    return getattr(ex, _SHOWN_ATTR, False)


def log_error(logger, ex, msg, *msg_args):
    if is_shown(ex):
        logger.error(msg, *msg_args)
    else:
        logger.exception(msg, *msg_args)



# © Copyright Databand.ai, an IBM Company 2022

from targets.target_config import TargetConfig, parse_target_config
from targets.values import ValueType


class TargetConfigValueType(ValueType):  # extracted for generics support
    """
    ValueType whose value is a ```target```.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
            some_input = TargetValueType()

            def run(self):
                  self.some_input.open('r')

    """

    type = TargetConfig

    support_from_str = True

    def parse_from_str(self, x):
        return parse_target_config(x)



# © Copyright Databand.ai, an IBM Company 2022

from tzlocal import get_localzone

from dbnd._core.utils.timezone import make_aware
from dbnd._vendor import click


class TZAwareDateTime(click.DateTime):
    name = "tzawaredatetime"

    def _try_to_convert_date(self, value, format):
        dt = super(TZAwareDateTime, self)._try_to_convert_date(value, format)
        if dt:
            return make_aware(dt, get_localzone())

    def __repr__(self):
        return "TZAwareDateTime"



# © Copyright Databand.ai, an IBM Company 2022

import json

from typing import List

import yaml

from six.moves import cPickle as pickle

from targets.marshalling.marshaller import Marshaller
from targets.target_config import FileFormat


class StrMarshaller(Marshaller):
    type = str
    file_format = FileFormat.txt

    clears_types_to_str = True

    def target_to_value(self, target, mode="r"):
        with target.open(mode) as fp:
            return fp.read()

    def value_to_target(self, value, target, mode="w"):
        if value is None:
            value = ""
        with target.open(mode) as fp:
            fp.write(value)
            fp.flush()


class StrLinesMarshaller(Marshaller):
    type = List[str]
    file_format = FileFormat.txt

    support_directory_direct_read = True
    support_multi_target_direct_read = True
    support_directory_direct_write = False
    clears_types_to_str = True

    def target_to_value(self, target, mode="r"):
        with target.open(mode) as fp:
            return fp.readlines()

    def value_to_target(self, value, target, mode="w", newline=True):
        if value is None:
            value = ""
        elif newline:
            value = ("%s\n" % s for s in value)
        with target.open(mode) as fp:
            fp.writelines(str(v) for v in value)
            fp.flush()
        return target


class ObjPickleMarshaller(Marshaller):
    type = object
    file_format = FileFormat.pickle

    def target_to_value(self, target, **kwargs):
        with target.open("rb") as fp:
            return pickle.load(fp, **kwargs)

    def value_to_target(self, value, target, **kwargs):
        with target.open("wb") as fp:
            pickle.dump(value, fp, **kwargs)
        return target


class ObjJsonMarshaller(Marshaller):
    type = object
    file_format = FileFormat.json

    def target_to_value(self, target, **kwargs):
        with target.open("r") as fp:
            return json.load(fp=fp, **kwargs)

    def value_to_target(self, value, target, **kwargs):
        from dbnd._core.utils.json_utils import json_default

        kwargs.setdefault("default", json_default)
        with target.open("w") as fp:
            json.dump(obj=value, fp=fp, **kwargs)
        return target


class ObjYamlMarshaller(Marshaller):
    type = object
    file_format = FileFormat.yaml

    def target_to_value(self, target, **kwargs):
        with target.open("r") as fp:
            return yaml.load(stream=fp, **kwargs)

    def value_to_target(self, value, target, **kwargs):
        from dbnd._core.utils.json_utils import json_default

        kwargs.setdefault("default", json_default)
        with target.open("w") as fp:
            yaml.dump(data=value, stream=fp, **kwargs)
        return target



# © Copyright Databand.ai, an IBM Company 2022


class CmdLineBuilder(object):
    def __init__(self):
        self._cmd_line = []

    def add(self, *args):
        self._cmd_line.extend(args)

    def extend(self, args):
        self._cmd_line.extend(args)

    def option(self, key, value):
        if value is None:
            return
        self.add(str(key), str(value))

    def option_bool(self, key, value):
        if value:
            self.add(str(key))

    def option_dict(self, key, value):
        pass

    def get_cmd(self):
        return list(self._cmd_line)

    def get_cmd_line(self, safe_curly_brackets=False):
        return list2cmdline(self._cmd_line, safe_curly_brackets=safe_curly_brackets)


def list2cmdline(seq, safe_curly_brackets=False):
    """
    Translate a sequence of arguments into a command line
    string, using the same rules as the MS C runtime:

    1) Arguments are delimited by white space, which is either a
       space or a tab.

    2) A string surrounded by double quotation marks is
       interpreted as a single argument, regardless of white space
       contained within.  A quoted string can be embedded in an
       argument.

    3) A double quotation mark preceded by a backslash is
       interpreted as a literal double quotation mark.

    4) Backslashes are interpreted literally, unless they
       immediately precede a double quotation mark.

    5) If backslashes immediately precede a double quotation mark,
       every pair of backslashes is interpreted as a literal
       backslash.  If the number of backslashes is odd, the last
       backslash escapes the next double quotation mark as
       described in rule 3.
    """

    # Based on subprocess.list2cmd
    # See
    # http://msdn.microsoft.com/en-us/library/17w5ykft.aspx
    # or search http://msdn.microsoft.com for
    # "Parsing C++ Command-Line Arguments"
    result = []
    needquote = False
    for arg in seq:
        bs_buf = []

        # Add a space to separate this argument from the others
        if result:
            result.append(" ")

        # FIX, support brackets
        # needquote = (" " in arg) or ("\t" in arg) or not arg
        needquote = (
            (" " in arg)
            or ("\t" in arg)
            or not arg
            or (safe_curly_brackets and "{" in arg)
        )
        if needquote:
            result.append('"')

        for c in arg:
            if c == "\\":
                # Don't know if we need to double yet.
                bs_buf.append(c)
            elif c == '"':
                # Double backslashes.
                result.append("\\" * len(bs_buf) * 2)
                bs_buf = []
                result.append('\\"')
            else:
                # Normal char
                if bs_buf:
                    result.extend(bs_buf)
                    bs_buf = []
                result.append(c)

        # Add remaining backslashes, if any.
        if bs_buf:
            result.extend(bs_buf)

        if needquote:
            result.extend(bs_buf)
            result.append('"')

    return "".join(result)


def list2cmdline_safe(seq, safe_curly_brackets=True):
    return list2cmdline(seq, safe_curly_brackets=safe_curly_brackets)



# © Copyright Databand.ai, an IBM Company 2022

import enum

from typing import Optional

import dbnd

from dbnd._core.utils.basics.singleton_context import SingletonContext


class TaskContextPhase(enum.Enum):
    BUILD = 0
    RUN = 1


class TaskContext(SingletonContext):
    def __init__(self, stack, phase):
        super(TaskContext, self).__init__()
        self.stack = stack
        self.phase = phase

    def __repr__(self):
        return "TaskContext( [...%s], %s)" % (self.stack[-1], self.phase)


def task_context(task, phase):
    # type: (dbnd.tasks.Task, TaskContextPhase) -> TaskContext
    base_stack = []

    if has_current_task():
        base_stack = current_task_stack()

    return TaskContext.new_context(
        stack=base_stack + [task], phase=phase, allow_override=True
    )


def has_current_task():
    return TaskContext.has_instance()


def current_task():
    # type: () -> dbnd.tasks.Task
    return TaskContext.get_instance().stack[-1]


def try_get_current_task():
    # type: () -> Optional[dbnd.tasks.Task]
    tc = TaskContext.try_get_instance()
    if tc and tc.stack:
        return tc.stack[-1]
    return None


def current():
    """
    Returns the current task's object.

    Example::

        from dbnd import current, task

        @task
        def calculate_alpha(alpha: int = 0.5):
            return current().task_version
    """
    return current_task()


def current_phase():
    # type: () -> TaskContextPhase
    return TaskContext.get_instance().phase


def current_task_stack():
    # type: () -> {}
    return TaskContext.get_instance().stack



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any, Dict, Optional, Type, Union

import six

from dbnd._core.current import try_get_databand_run
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._core.task.base_task import _BaseTask


def get_remote_engine_name():
    # type: () -> Optional[str]
    """Retrieve the name of the remote engine of the current run if exists, None otherwise."""
    run = try_get_databand_run()
    if run:
        return run.run_executor.remote_engine.task_name


def get_task_params_values(task):
    # type: (_BaseTask) -> Dict[str, Any]
    """Access the task's user parameters values as a map between the name of the param and its current value."""
    return {
        param.name: param.value
        for param in task.task_params.get_param_values(ParameterFilters.USER)
    }


def get_task_params_defs(task):
    # type: (Union[_BaseTask, Type[_BaseTask]]) -> Dict[str, ParameterDefinition]
    """
    Access the task's parameter definitions as a map between the name of the param and the parameter definition.

    Works with both task and task_class.
    """
    return {
        k: v
        for k, v in six.iteritems(task.task_definition.task_param_defs)
        if not v.system
    }



# © Copyright Databand.ai, an IBM Company 2022

from targets.pipes.base import (
    InputPipeProcessWrapper,
    IOPipeline,
    OutputPipeProcessWrapper,
)


class GzipPipeline(IOPipeline):

    input = "bytes"
    output = "bytes"

    def __init__(self, compression_level=None):
        self.compression_level = compression_level

    def pipe_reader(self, input_pipe):
        return InputPipeProcessWrapper(["gunzip"], input_pipe)

    def pipe_writer(self, output_pipe):
        args = ["gzip"]
        if self.compression_level is not None:
            args.append("-" + str(int(self.compression_level)))
        return OutputPipeProcessWrapper(args, output_pipe)



# © Copyright Databand.ai, an IBM Company 2022

import re

from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import attr

from sqlparse.sql import (
    Comparison,
    Function,
    Identifier,
    IdentifierList,
    Parenthesis,
    Token,
    TokenList,
)
from sqlparse.tokens import Keyword, Name, Whitespace, Wildcard

from dbnd._core.constants import DbndTargetOperationType
from dbnd._core.utils.sql_tracker_common.utils import ddict2dict, strip_quotes


@attr.s
class Column:
    dataset_name = attr.ib()
    name = attr.ib(default=None)
    alias = attr.ib(default=None)
    is_file = attr.ib(default=False)
    is_stage = attr.ib(default=False)

    @property
    def is_wildcard(self):
        return self.name == "*"


Columns = List[Column]
Schema = Dict[str, Columns]

OP_TYPE = DbndTargetOperationType
READ = DbndTargetOperationType.read
WRITE = DbndTargetOperationType.write
READ_OPERATIONS = {"FROM"}
CLOUD_SERVICE_URI_REGEX = re.compile(r"s3://\S+|azure://\S+|gcs://\S+")

STAGE_REGEX = re.compile(r"@\S+")


def calculate_file_path_default(file_path):
    return file_path


class SqlQueryExtractor:
    def __init__(self, calculate_file_path=calculate_file_path_default):
        self.calculate_file_path = calculate_file_path

    def extract_operations_schemas(self, statement: TokenList) -> Dict[OP_TYPE, Schema]:
        """
        This function go over the statement and extract the used tables names and the operation that it is been done
        with them. The result is a list of pairs where the left element is the table name and the right one is the
        operation.
        """
        # `sqlparse` implementation details:
        # Using the changing `idx` (the index of the current token) to keep moving in the token list (statement),
        # after finding a current token. First idx is -1 because this is the first position outside the token list.
        # Each call to `token_next` or `token_next_by` using the current idx to start searching from the current
        # position
        idx = -1

        # holds the columns we find in the scope and not yet assigned to any schema
        columns: Columns = []
        # holds any dynamic table schema that is created in the scope.
        # dynamic tables are created by CTE statements
        dynamic_tables: Dict[str, Schema] = {}

        # collect the operations and their schemas
        operations = defaultdict(lambda: defaultdict(list))
        is_copy_into_handled = False

        while True:
            # each iteration we pick up the next token we didn't handle yet
            idx, token = self._next_non_empty_token(idx, statement)
            if not token:
                # done iterating over the sql statement
                # converting the result to regular dict
                return ddict2dict(operations)

            token: Token
            if token.ttype in Keyword:
                operation_name = token.value.upper()
                idx_potential, next_token = self._next_non_empty_token(idx, statement)
                next_operation_name = next_token.value.upper() if next_token else None
                # copy and copy into statement are handled once per query
                if (
                    operation_name == "INTO"
                    or (operation_name == "COPY" and next_operation_name != "INTO")
                    and not is_copy_into_handled
                ):
                    extracted, idx = self.handle_into(
                        idx_potential, next_token, statement
                    )
                    operations[WRITE].update(extracted)
                    # There could be few `COPY` tokens, we're processing only the first one
                    is_copy_into_handled = True

                elif operation_name in READ_OPERATIONS:
                    extracted, idx, columns = self.handle_read_operation(
                        idx_potential, next_token, columns, dynamic_tables
                    )
                    # to support UNION we extend the array of the same column
                    for k, v in extracted.items():
                        operations[READ][k].extend(v)

                elif token.ttype is Keyword.CTE:
                    # after a CTE statement (WITH) we expect identifiers list with each has a sub-query inside parenthesis
                    idx, token = self._next_non_empty_token(idx, statement)

                    if isinstance(token, IdentifierList):
                        token_list = token.get_identifiers()
                    else:
                        token_list = [token]

                    for identifier in token_list:
                        _, cte_statement = identifier.token_next_by(i=Parenthesis)
                        cte_operations = self.extract_operations_schemas(cte_statement)
                        extracted = self.enrich_with_dynamic(
                            cte_operations[READ], dynamic_tables
                        )
                        dynamic_tables[identifier.get_name()] = extracted

    def handle_read_operation(
        self,
        idx: int,
        next_token: Token,
        columns: Columns,
        dynamic_tables: Dict[str, Schema],
    ) -> Tuple[Schema, int, Columns]:
        extracted = {}
        if isinstance(next_token, Parenthesis):
            return self.handle_nested_query(next_token)
        if next_token.ttype not in Keyword and isinstance(next_token, Identifier):
            # no subquery - just parse the source/dest name of the operator
            extracted, left_columns = self.generate_schema(columns, next_token)
            columns = left_columns
            extracted = self.enrich_with_dynamic(extracted, dynamic_tables)

        return extracted, idx, columns

    def handle_nested_query(self, nested_statement):
        columns = []
        is_select_query = False
        nested_idx = 0
        for token in nested_statement.tokens:
            # identify nested query is select
            if token.normalized == "SELECT":
                is_select_query = True
            # only select nested queries are supported
            elif is_select_query:
                # TODO: we might want to parse query columns for nested read operations
                # if isinstance(token, IdentifierList):
                # columns.append(Column(dataset_name='',name=token.normalized, alias=token.normalized))
                if token.normalized in READ_OPERATIONS:
                    next = self._next_non_empty_token(nested_idx, nested_statement)
                    if not next:
                        continue
                    extracted, nested_idx, columns = self.handle_read_operation(
                        nested_idx, next[1], columns, {}
                    )
            nested_idx += 1
        return extracted, nested_idx, columns

    def clean_query(self, query: str) -> str:
        """
        Cleans apostrophe and quote characters from cloud url or stage in query to parse it successfully by sqlparse
         :param str query: query statement
        """

        command_list = query.split(" ")
        clean_query_list = []
        for command in command_list:
            if command and (self.find_cloud_regex(command) or self.is_stage(command)):
                command = strip_quotes(command)
            clean_query_list.append(command)
        return " ".join(clean_query_list).strip()

    def is_stage(self, table_name: str) -> str:
        """
        Returns a pattern of file stage if exits
        :param str table_name: name of a table

        """
        if table_name:
            return STAGE_REGEX.findall(table_name.split(".")[-1])

    def generate_schema(
        self, columns: Columns, next_token: Identifier
    ) -> Tuple[Schema, Columns]:
        """
        generate schema object and columns for table, file or stage

        :param Columns columns: name of a table we want to track
        :param Identifier next_token: query identifier

        """
        cloud_uri_path = self.extract_cloud_uri(next_token)
        if cloud_uri_path:
            table_alias = self.calculate_file_path(cloud_uri_path)
            table_name = table_alias
            columns = [
                Column(
                    dataset_name=table_name, name="*", alias=table_alias, is_file=True
                )
            ]
        else:
            table_alias = self.get_identifier_name(next_token)
            table_name = self.get_full_name(next_token)
            if self.is_stage(table_name):
                columns = [
                    Column(
                        dataset_name=table_name,
                        name="*",
                        alias=table_alias,
                        is_stage=True,
                    )
                ]
        extracted, left_columns = self.extract_schema(table_name, table_alias, columns)
        return extracted, left_columns

    def handle_into(
        self, idx: int, next_token: Token, statement: TokenList
    ) -> Tuple[Schema, int]:
        if next_token.is_group:
            # if token is group, get inner function token if exist
            for token in next_token.tokens:
                if isinstance(token, Function):
                    next_token = token
        table_name = self.get_identifier_name(next_token)
        table_full_name = self.get_full_name(next_token)

        if isinstance(next_token, Function):
            extracted = self.extract_write_schema(
                next_token.get_parameters(), table_name, table_name
            )
        else:
            extracted = self.extract_write_schema(
                [Token(Wildcard, "*")], table_full_name, table_name
            )
        return extracted, idx

    def extract_write_schema(
        self, tokens: List[Token], table_name: str, table_alias: str
    ) -> Schema:
        cols = self._extract_columns(tokens)
        extracted, _ = self.extract_schema(table_name, table_alias, cols)
        return extracted

    def extract_schema(
        self, table_name: str, table_alias: str, columns: Columns
    ) -> Tuple[Schema, Columns]:
        """
        Collects a schema and a wildcard column for a table, file or stage
        :param str table_name: full path and name of a table , file or stage
        :param str table_alias: alias of table, file or stage
        :param Columns columns: array of columns for a table, file or stage (for file and stage would be wildcard)
        """
        collected_schema = defaultdict(list)
        left_columns = []

        col: Column
        for col in columns:
            if col.is_wildcard:
                ev_col = attr.evolve(
                    col, dataset_name=table_name, alias=f"{table_name}.*"
                )
                collected_schema[ev_col.alias].append(ev_col)
                left_columns.append(col)

            elif col.dataset_name and col.dataset_name != table_alias:
                left_columns.append(col)

            else:
                ev_col = attr.evolve(col, dataset_name=table_name)
                collected_schema[ev_col.alias].append(ev_col)

        return collected_schema, left_columns

    @staticmethod
    def _next_non_empty_token(
        idx: int, token_list: TokenList
    ) -> Tuple[Optional[int], Optional[Token]]:
        return token_list.token_next(idx, skip_ws=True, skip_cm=True)

    def enrich_with_dynamic(
        self, extracted: Schema, dynamic: Dict[str, Schema]
    ) -> Schema:
        if not dynamic:
            return extracted

        result = defaultdict(list)
        for name, cols in extracted.items():
            for col in cols:
                if col.dataset_name in dynamic:
                    dynamic_table_schema = dynamic[col.dataset_name]
                    if col.is_wildcard:
                        result.update(dynamic_table_schema)

                    elif col.name in dynamic_table_schema:
                        result[col.name].extend(dynamic_table_schema[col.name])

                    else:
                        result[col.name].append(col)
                else:
                    result[name].append(col)
        return result

    def _extract_columns(self, tokens: List[Token]) -> Columns:
        collected_columns = []
        for token in tokens:
            if token.ttype == Wildcard:
                col = Column(dataset_name=None, name="*", alias=None)

            elif isinstance(token, Identifier):
                name = token.get_real_name()
                alias = token.get_alias() or name

                first = token.token_first()
                ref = (
                    first.value
                    if len(token.tokens) > 1 and first.ttype == Name
                    else None
                )
                col = Column(dataset_name=ref, name=name, alias=alias)

            elif isinstance(token, Comparison):
                name = token.left.get_real_name()
                col = Column(dataset_name=None, name=name, alias=name)

            else:
                name = token.normalized or token.get_alias()
                col = Column(dataset_name=None, name=name, alias=name)

            collected_columns.append(col)
        return collected_columns

    @staticmethod
    def get_identifier_name(identifier):
        return identifier.get_alias() or identifier.value.split(".")[-1]

    def extract_cloud_uri(self, identifier: Identifier) -> str:
        """
        Search for cloud providers pattern in query identifier,
        if pattern exists returns the URI as string

        :param Identifier identifier: query identifier
        """
        if identifier.value.lower() in ["s3", "azure", "gcs"]:
            return self.find_cloud_regex(identifier.parent.value)

    def find_cloud_regex(self, identifier: str) -> str:
        cloud_uri = CLOUD_SERVICE_URI_REGEX.findall(identifier.lower())
        return cloud_uri[0] if cloud_uri else None

    @staticmethod
    def get_full_name(identifier):
        # "name as alias" or "name alias" or "complicated column expression alias"
        ws_index, ws = identifier.token_next_by(t=Whitespace)
        if len(identifier.tokens) > 2 and ws is not None:
            return "".join(token.value for token in identifier.tokens[:ws_index])

        return identifier.value



# © Copyright Databand.ai, an IBM Company 2022

from collections import defaultdict
from typing import Union


def ddict2dict(d: Union[defaultdict, dict]) -> dict:
    """convert defaultdict into a dict"""
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = ddict2dict(v)
    return dict(d)


def strip_quotes(command: str) -> str:
    if command:
        return command.replace("'", "").replace('"', "")


def get_redshift_uri(host: str, db: str, schema: str, table: str) -> str:
    return f"redshift://{host}/{db}/{schema}/{table}".lower()



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging

from dbnd._core.constants import AlertDefOperator, AlertSeverity, RunState, TaskRunState
from dbnd._core.errors.base import DatabandApiError
from dbnd._core.utils.basics.text_banner import TextBanner, safe_tabulate
from dbnd._core.utils.cli import options_dependency, required_mutually_exclusive_options
from dbnd._vendor import click
from dbnd.api.alerts import (
    create_alert,
    delete_alerts,
    get_alerts_filtered,
    list_job_alerts,
)
from dbnd.cli.cmd_alerts_constants import (
    ANOMALY_ALERT_LOOK_BACK_HELP,
    ANOMALY_ALERT_SENSITIVITY_HELP,
    CREATE_HELP_MSG,
    DELETE_HELP_MSG,
    IS_STR_VALUE_HELP,
    LIST_HELP_MSG,
    METRIC_HELP,
    RANGE_ALERT_BASE_LINE_HELP,
    RANGE_ALERT_RANGE_HELP,
)


logger = logging.getLogger(__name__)


class Severity(click.Choice):
    def __init__(self):
        super(Severity, self).__init__(
            choices=AlertSeverity.values(), case_sensitive=False
        )

    def convert(self, value, param, ctx):
        return super(Severity, self).convert(value.upper(), param, ctx)


class RunStateChoice(click.Choice):
    def __init__(self):
        super(RunStateChoice, self).__init__(
            choices=RunState.all_values(), case_sensitive=False
        )

    def convert(self, value, param, ctx):
        return super(RunStateChoice, self).convert(value.upper(), param, ctx)


class TaskRunStateChoice(click.Choice):
    def __init__(self):
        super(TaskRunStateChoice, self).__init__(
            choices=TaskRunState.all_values(), case_sensitive=False
        )

    def convert(self, value, param, ctx):
        return super(TaskRunStateChoice, self).convert(value.upper(), param, ctx)


operator_to_required_options = {
    AlertDefOperator.RANGE.value: ["baseline", "range"],
    AlertDefOperator.ANOMALY.value: ["look_back", "sensitivity"],
    AlertDefOperator.EQUAL.value: ["value"],
    AlertDefOperator.NOT_EQUAL.value: ["value"],
    AlertDefOperator.GREATER_THAN.value: ["value"],
    AlertDefOperator.NOT_LESS_THAN.value: ["value"],
    AlertDefOperator.LESS_THAN.value: ["value"],
    AlertDefOperator.NOT_GREATER_THAN.value: ["value"],
}

supported_alerts = {
    "run_duration": dict(
        value_help="Run duration of pipeline in seconds",
        value_type=click.INT,
        with_task=False,
        alert_type="RunDuration",
        operator=None,
        metric="Run Duration (seconds)",
        is_str_value=False,
    ),
    "run_state": dict(
        value_help="State of the run to alert on",
        value_type=RunStateChoice(),
        with_task=False,
        alert_type="RunState",
        operator="==",
        metric="Run State",
        is_str_value=False,
    ),
    "ran_last_x_seconds": dict(
        value_help="Duration in seconds of missing success run from pipeline",
        value_type=click.INT,
        with_task=False,
        alert_type="RanLastXSeconds",
        operator=">",
        metric="Delay Between Subsequent Runs (seconds)",
        is_str_value=False,
    ),
    "task_state": dict(
        value_help="State of the task run to alert on",
        value_type=TaskRunStateChoice(),
        with_task=True,
        alert_type="TaskState",
        operator="==",
        metric="State",
        is_str_value=True,
    ),
    "custom_metric": dict(
        value_help="Threshold to match the metric",
        value_type=click.STRING,
        with_task=True,
        alert_type="CustomMetric",
        operator=None,
        metric=None,
        is_str_value=None,
    ),
}


def get_option(ctx, param_name: str, is_parent_command: bool):
    parent_params = (
        ctx.parent.command.params if is_parent_command else ctx.command.params
    )
    return next(op for op in parent_params if op.name == param_name)


def validate_option(
    ctx,
    option_name: str,
    option_value: str,
    option_condition: bool,
    from_parent_command: bool,
):
    if not option_condition and option_value:
        raise click.BadParameter(
            f"'{option_name}' is set but it is unavailable for this alert",
            param=get_option(ctx, option_name, from_parent_command),
            ctx=ctx,
        )
    if option_condition and not option_value:
        raise click.MissingParameter(
            ctx=ctx,
            param=get_option(ctx, option_name, from_parent_command),
            message=f"'{option_name}' is required for this alert",
        )


def validate_passed_options(ctx, with_task, operator, str_value, **more_options):
    validate_option(ctx, "task", ctx.obj.get("task"), with_task, True)
    # support only equal\not equal operator with task of str value
    if (
        with_task
        and str_value
        and operator
        not in [AlertDefOperator.EQUAL.value, AlertDefOperator.NOT_EQUAL.value]
    ):
        raise click.BadArgumentUsage(
            f"Metrics with string values can be operated only with '==' and '!=', "
            f"But operator '{operator}' was supplied.",
            ctx=ctx,
        )
    for option_name, option_value in more_options.items():
        is_required_option = option_name in operator_to_required_options[operator]
        validate_option(ctx, option_name, option_value, is_required_option, False)


def bind_create_to_supported_alert(
    name, alert_type, with_task, value_help, value_type, operator, metric, is_str_value
):
    @click.option("--value", "-v", help=value_help, type=value_type, required=False)
    @click.option(
        "--baseline",
        "-bl",
        help=RANGE_ALERT_BASE_LINE_HELP,
        type=click.FLOAT,
        required=False,
    )
    @click.option(
        "--range", "-r", help=RANGE_ALERT_RANGE_HELP, type=click.FLOAT, required=False
    )
    @click.option(
        "--look-back",
        "-lb",
        help=ANOMALY_ALERT_LOOK_BACK_HELP,
        type=click.INT,
        required=False,
    )
    @click.option(
        "--sensitivity",
        "-sn",
        help=ANOMALY_ALERT_SENSITIVITY_HELP,
        type=click.INT,
        required=False,
    )
    @click.pass_context
    def create_alert_command(
        ctx,
        value,
        baseline,
        range,
        look_back,
        sensitivity,
        metric_name=metric,
        op=operator,
        str_value=is_str_value,
    ):
        validate_passed_options(
            ctx,
            with_task,
            op,
            str_value,
            value=value,
            baseline=baseline,
            range=range,
            look_back=look_back,
            sensitivity=sensitivity,
        )
        more_options = {
            "baseline": baseline,
            "range": range,
            "look_back": look_back,
            "sensitivity": sensitivity,
        }
        cmd_create_alert(
            manage_ctx=ctx,
            alert_type=alert_type,
            operator=op,
            value=value,
            user_metric=metric_name,
            str_value=str_value,
            **more_options,
        )

    if operator is None:
        create_alert_command = click.option(
            "--op",
            "-o",
            help="your operator",
            type=click.Choice(AlertDefOperator.all_values()),
            required=True,
        )(create_alert_command)

    if metric is None:
        create_alert_command = click.option(
            "--metric-name", "-m", help=METRIC_HELP, type=click.STRING, required=True
        )(create_alert_command)

    if is_str_value is None:
        create_alert_command = click.option(
            "--str-value", help=IS_STR_VALUE_HELP, type=click.BOOL, required=True
        )(create_alert_command)

    create_alert_command.__name__ = name
    return create_alert_command


@click.group()
@click.pass_context
def alerts(ctx):
    """Manage Databand alerts"""


@alerts.group(help=CREATE_HELP_MSG)
@click.option("--severity", "-s", help="Alert severity", type=Severity(), required=True)
@click.option(
    "--pipeline", "-p", help="Pipeline name", type=click.STRING, required=False
)
@click.option("--pipeline-id", type=click.STRING, required=False)
@click.option("--task", "-t", help="Task name", type=click.STRING, required=False)
@click.option(
    "--update",
    "-u",
    "uid",
    help="Uid of an existing alert to update",
    type=click.STRING,
    required=False,
)
@click.pass_context
def create(ctx, pipeline, pipeline_id, severity, task, uid):
    """Create or update alerts for given job."""
    ctx.obj = {
        "job": pipeline,
        "job_id": pipeline_id,
        "severity": severity,
        "task": task,
        "uid": uid,
    }


@alerts.command("list", help=LIST_HELP_MSG)
@click.option(
    "--pipeline", "-p", help="Pipeline name", type=click.STRING, required=True
)
def list_jobs(pipeline):
    try:
        alerts_list = list_job_alerts(job_name=pipeline)
    except LookupError as e:
        logger.warning(e)
    else:
        print_table("Alerts configured for {job}".format(job=pipeline), alerts_list)
        alert_uids = [alert["uid"] for alert in alerts_list]
        logger.info("Alerts listed: %s", str(alert_uids))


@alerts.command(help=DELETE_HELP_MSG)
@options_dependency("pipeline", "name")
@required_mutually_exclusive_options("uid", "wipe", "pipeline")
@click.option("--uid", "-u", help="alert uid", type=click.STRING)
@click.option("--wipe", help="delete all alerts", is_flag=True)
@click.option("--pipeline", "-p", help="pipeline name", type=click.STRING)
@click.option("--name", "-n", help="alert custom name", type=click.STRING)
def delete(uid, wipe, pipeline, name):
    if wipe:
        alerts_list = get_alerts_filtered()
    elif uid:
        alerts_list = get_alerts_filtered(alert_def_uid=uid)
    else:
        alerts_list = get_alerts_filtered(job_name=pipeline, custom_name=name)
    if alerts_list:
        uids = [alert["uid"] for alert in alerts_list]
        delete_alerts(uids)
        print_table("deleted", alerts_list)
        logger.info("Deleted uids: %s", str(uids))
    else:
        logger.warning("Alerts not found to delete")


def print_table(header, alerts_list):
    banner = TextBanner(header)
    banner.write(build_alerts_table(alerts_list))
    logger.info(banner.get_banner_str())


def build_alerts_table(alerts_data):
    extract_keys = (
        "uid",
        "job_name",
        "task_name",
        "task_repr",
        "custom_name",
        "severity",
        "type",
        "operator",
        "value",
        "user_metric",
    )
    headers = (
        "uid",
        "pipeline",
        "task",
        "task_repr",
        "name",
        "severity",
        "type",
        "op",
        "value",
        "metric",
    )
    table_data = []
    for alert in alerts_data:
        alert_row = [alert.get(key, "") for key in extract_keys]
        table_data.append(alert_row)
    return safe_tabulate(table_data, headers)


def assign_operator_to_alert_type(base_alert_type, operator):
    alert_type = base_alert_type
    if operator == AlertDefOperator.ANOMALY.value:
        alert_type = "ML" + alert_type
    if operator == AlertDefOperator.RANGE.value:
        alert_type += "Range"
    alert_type += "Alert"
    return alert_type


def cmd_create_alert(
    manage_ctx, alert_type, operator, value, user_metric, str_value, **more_options
):
    full_alert_type = assign_operator_to_alert_type(alert_type, operator)
    task = manage_ctx.obj.get("task", None)  # Currently used in task_name & task_repr
    try:
        alert_def_uid = create_alert(
            job_name=manage_ctx.obj.get("job"),
            job_id=manage_ctx.obj.get("job_id"),
            task_name=task,
            task_repr=task,
            uid=manage_ctx.obj.get("uid", None),
            alert_class=full_alert_type,
            severity=manage_ctx.obj["severity"],
            operator=operator,
            value=value,
            user_metric=user_metric,
            is_str_value=str_value,
            **more_options,
        )
    except DatabandApiError as e:
        logger.warning("failed with - %s", e.response)
    else:
        logger.info("Created alert %s", alert_def_uid["uid"])


def register_create_command_to_alerts():
    # Registering the create commands
    for name, values in supported_alerts.items():
        binder_to_sub_command = bind_create_to_supported_alert(name, **values)
        create.command(name=name.replace("_", "-"))(binder_to_sub_command)


register_create_command_to_alerts()



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import typing

from abc import ABCMeta, abstractmethod
from itertools import chain

import six

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.current import get_databand_run
from dbnd._core.log.logging_utils import TaskContextFilter
from dbnd._core.task_build.task_context import task_context
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.traversing import traverse_to_str


if typing.TYPE_CHECKING:
    from typing import Any, Dict, Optional

    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.parameter.parameter_value import Parameters
    from dbnd._core.settings import DatabandSettings, EnvConfig
    from dbnd._core.task_ctrl.task_dag import _TaskDagNode
    from dbnd._core.task_ctrl.task_relations import TaskRelations
    from dbnd._core.task_ctrl.task_validator import TaskValidator
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


class TaskSubCtrl(object):
    def __init__(self, task):
        super(TaskSubCtrl, self).__init__()

        from dbnd._core.task.task import Task

        self.task = task  # type: Task

    @property
    def dbnd_context(self):
        # type: (TaskSubCtrl) -> DatabandContext
        return self.task.dbnd_context

    @property
    def ctrl(self):
        # type: (TaskSubCtrl) -> TaskCtrl
        return self.task.ctrl

    @property
    def task_id(self):
        # type: (TaskSubCtrl) -> str
        return self.task.task_id

    @property
    def task_dag(self):
        # type: (TaskSubCtrl) -> _TaskDagNode
        return self.ctrl._task_dag

    @property
    def visualiser(self):
        return self.ctrl._visualiser

    @property
    def meta_info_serializer(self):
        return self.ctrl._meta_info_serializer

    @property
    def params(self):  # type: () -> Parameters
        return self.task._params

    @property
    def settings(self):  # type: ()-> DatabandSettings
        return self.task.settings

    @property
    def relations(self):  # type: () -> TaskRelations
        return self.ctrl._relations

    @property
    def validator(self):  # type: () -> TaskValidator
        return self.ctrl.task_validator

    @property
    def task_env(self):
        # type: ()-> EnvConfig
        return self.task.task_env

    def get_task_by_task_id(self, task_id):
        return self.dbnd_context.task_instance_cache.get_task_by_id(task_id)

    def __repr__(self):
        return "%s.%s" % (self.task.task_id, self.__class__.__name__)


@six.add_metaclass(ABCMeta)
class _BaseTaskCtrl(TaskSubCtrl):
    def __init__(self, task):
        super(_BaseTaskCtrl, self).__init__(task)

        from dbnd._core.task_ctrl.task_dag import _TaskDagNode  # noqa: F811
        from dbnd._core.task_ctrl.task_dag_describe import DescribeDagCtrl
        from dbnd._core.task_ctrl.task_descendant import (
            TaskDescendants as _TaskDescendants,
        )
        from dbnd._core.task_ctrl.task_repr import TaskRepr
        from dbnd._core.task_ctrl.task_visualiser import TaskVisualiser  # noqa: F811

        self._task_dag = _TaskDagNode(task)
        self.descendants = _TaskDescendants(task)

        self._visualiser = TaskVisualiser(task)
        self.describe_dag = DescribeDagCtrl(task)

        self.task_repr = TaskRepr(self.task)

        # will be assigned by the latest Run
        self.last_task_run = None  # type: Optional[TaskRun]
        self.force_task_run_uid = None  # force task run uid

    def _initialize_task(self):
        # only at the end we can build the final version of "function call"
        self.task_repr.initialize()

    def banner(self, msg, color=None, task_run=None):
        return self.visualiser.banner(msg=msg, color=color, task_run=task_run)

    @property
    def task_run(self):
        # type: ()-> TaskRun
        run = get_databand_run()
        return run.get_task_run(self.task.task_id)

    @contextlib.contextmanager
    def task_context(self, phase):
        # we don't want logs/user wrappers at this stage
        with nested(
            task_context(self.task, phase),
            TaskContextFilter.task_context(self.task.task_id),
        ):
            yield

    @abstractmethod
    def should_run(self):
        pass

    def io_params(self):
        return chain(self.task_outputs.values(), self.task_inputs.values())

    @property
    @abstractmethod
    def task_inputs(self):
        # type: () -> Dict[str, Dict[str][Any]]
        raise NotImplementedError()

    @property
    @abstractmethod
    def task_outputs(self):
        # type: () -> Dict[str, Dict[str][Any]]
        raise NotImplementedError()


class TaskCtrl(_BaseTaskCtrl):
    def __init__(self, task):
        from dbnd._core.task.task import Task

        assert isinstance(task, Task)

        super(TaskCtrl, self).__init__(task)

        from dbnd._core.task_ctrl.task_relations import TaskRelations  # noqa: F811
        from dbnd._core.task_ctrl.task_validator import TaskValidator

        self._relations = TaskRelations(task)
        self.task_validator = TaskValidator(task)

        self._should_run = self.task._should_run()

    def _initialize_task(self):
        # target driven relations are relevant only for orchestration tasks
        self.relations.initialize_relations()
        self.task_dag.initialize_dag_node()

        super(TaskCtrl, self)._initialize_task()

        # validate circle dependencies
        # may be we should move it to global level because of performance issues
        # however, by running it at every task we'll be able to find the code that causes the issue
        # and show it to user
        if self.dbnd_context.settings.run.recheck_circle_dependencies:
            self.task_dag.topological_sort()

    def should_run(self):
        # convert to property one day
        return self._should_run

    def subdag_tasks(self):
        return self.task_dag.subdag_tasks()

    def save_task_band(self):
        if self.task.task_band:
            task_outputs = traverse_to_str(self.task.task_outputs)
            self.task.task_band.as_object.write_json(task_outputs)

    @contextlib.contextmanager
    def task_context(self, phase):
        # we don't want logs/user wrappers at this stage
        with nested(
            task_context(self.task, phase),
            TaskContextFilter.task_context(self.task.task_id),
            config.config_layer_context(self.task.task_config_layer),
        ):
            yield

    @property
    def task_inputs(self):
        return self.relations.task_inputs

    @property
    def task_outputs(self):
        return self.relations.task_outputs


class TrackingTaskCtrl(_BaseTaskCtrl):
    def __init__(self, task):
        from dbnd._core.task.tracking_task import TrackingTask

        assert isinstance(task, TrackingTask)

        super(TrackingTaskCtrl, self).__init__(task)

    @property
    def task_inputs(self):
        return dict(dict())

    @property
    def task_outputs(self):
        return dict(dict())

    def should_run(self):
        return True



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.parameter import PARAMETER_FACTORY as parameter
from dbnd._core.task import Config


class HistogramConfig(Config):
    _conf__task_family = "histogram"

    spark_parquet_cache_dir = parameter(
        default=None,
        description="Enables pre-cache DF using .parquet store at `spark_temp_dir`",
    )[str]

    spark_cache_dataframe = parameter(
        default=False, description="Enables caching of the whole frame"
    )[bool]

    spark_cache_dataframe_column = parameter(
        default=True,
        description="Enables caching of the numerical df during histogram calculation",
    )[str]



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import absolute_import

import numpy as np

from targets.marshalling.marshaller import Marshaller
from targets.target_config import FileFormat


class NumpyArrayMarshaller(Marshaller):
    type = np.ndarray
    file_format = FileFormat.numpy

    def target_to_value(self, target, **kwargs):
        """
        :param obj: object to pickle
        :return:
        """
        with target.open("rb") as fp:
            return np.load(fp, **kwargs)

    def value_to_target(self, value, target, **kwargs):
        """
        :param obj: object to pickle
        :return:
        """
        target.mkdir_parent()
        with target.open("wb") as fp:
            np.save(fp, value, **kwargs)


class NumpyArrayPickleMarshaler(NumpyArrayMarshaller):
    file_format = FileFormat.pickle

    def target_to_value(self, target, **kwargs):
        return np.load(target.path, allow_pickle=True, **kwargs)

    def value_to_target(self, value, target, **kwargs):
        np.save(target.path, value, allow_pickle=True, **kwargs)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from targets.values.builtins_values import (
    BoolValueType,
    CallableValueType,
    FloatValueType,
    IntValueType,
    NullableStrValueType,
    ObjectValueType,
    StrValueType,
    ValueType,
)
from targets.values.datetime_value import DateTimeValueType, DateValueType
from targets.values.registry import ValueTypeRegistry
from targets.values.structure import (
    DictValueType,
    ListValueType,
    SetValueType,
    TupleValueType,
)
from targets.values.target_config_value import TargetConfigValueType
from targets.values.target_values import (
    TargetPathLibValueType,
    TargetPathValueType,
    TargetValueType,
)
from targets.values.timedelta_value import DateIntervalValueType, TimeDeltaValueType
from targets.values.value_type import InlineValueType  # noqa: F401
from targets.values.version_value import VersionValueType


if typing.TYPE_CHECKING:
    from typing import Optional


logger = logging.getLogger(__name__)


known_values = []
try:
    import matplotlib  # noqa: F401

    from targets.values.matplotlib_values import MatplotlibFigureValueType

    known_values.append(MatplotlibFigureValueType())
except ImportError:
    pass

try:
    import numpy  # noqa: F401
    import pandas  # noqa: F401

    from targets.values.pandas_values import (
        DataFramesDictValueType,
        DataFrameValueType,
        PandasSeriesValueType,
    )

    known_values.append(DataFrameValueType())
    known_values.append(PandasSeriesValueType())
    known_values.append(DataFramesDictValueType())
except ImportError:
    pass

try:
    import numpy  # noqa: F401

    from targets.values.numpy_values import NumpyArrayValueType

    known_values.append(NumpyArrayValueType())
except ImportError:
    pass


# Note: order matters. Examples:
# isinstance(True, int) == True, so it's important to have bool check before int
# isinstance(datetime.datetime.utc(), date) == True
known_values.extend(
    [
        # simple
        BoolValueType(),
        IntValueType(),
        FloatValueType(),
        # date/time
        DateValueType(),
        DateTimeValueType(),
        TimeDeltaValueType(),
        DateIntervalValueType(),
        # structs,
        ListValueType(),
        DictValueType(),
        SetValueType(),
        TupleValueType(),
        # targets, path
        TargetValueType(),
        TargetPathValueType(),
        TargetPathLibValueType(),
        TargetConfigValueType(),
        CallableValueType(),
        # str types
        VersionValueType(),
        StrValueType(),
        NullableStrValueType(),
    ]
)


# OBJECT VALUE is always the last
known_values.append(ObjectValueType())

_VALUE_TYPES = ValueTypeRegistry(known_value_types=known_values)


def get_types_registry():
    # type: ()-> ValueTypeRegistry
    return _VALUE_TYPES


def get_value_type_of_obj(obj, default_value_type=None):
    # type: (object, Optional[ValueType]) -> Optional[ValueType]
    return get_types_registry().get_value_type_of_obj(obj, default_value_type)


def get_value_type_of_type(obj_type, inline_value_type=False):
    return get_types_registry().get_value_type_of_type(
        obj_type, inline_value_type=inline_value_type
    )


def register_value_type(value_type):
    return get_types_registry().register_value_type(value_type)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.tracking.metrics import TRACKER_MISSING_MESSAGE, _get_tracker
from dbnd._core.utils.one_time_logger import get_one_time_logger
from dbnd.utils.dbt_cloud_api_client import DbtCloudApiClient


logger = logging.getLogger(__name__)


def _report_dbt_metadata(dbt_metadata, tracker=None):
    if tracker is None:
        tracker = _get_tracker()

    if not tracker:
        message = TRACKER_MISSING_MESSAGE % ("report_dbt_metadata",)
        get_one_time_logger().log_once(message, "report_dbt_metadata", logging.WARNING)
        return

    tracker.log_dbt_metadata(dbt_metadata=dbt_metadata)


def get_run_data_from_dbt(dbt_cloud_client, dbt_job_run_id):
    dbt_run_meta_data = dbt_cloud_client.get_run(run_id=dbt_job_run_id)

    if not dbt_run_meta_data:
        logger.warning("Fail getting run data from dbt cloud ")
        return None

    env_id = dbt_run_meta_data.get("environment_id")
    env = dbt_cloud_client.get_environment(env_id=env_id)

    if env:
        dbt_run_meta_data["environment"] = env

    for step in dbt_run_meta_data.get("run_steps", []):
        step_run_results_artifact = dbt_cloud_client.get_run_results_artifact(
            run_id=dbt_job_run_id, step=step["index"]
        )
        if step_run_results_artifact:
            step["run_results"] = step_run_results_artifact

        step_run_manifest_artifact = dbt_cloud_client.get_manifest_artifact(
            run_id=dbt_job_run_id, step=step["index"]
        )

        if step_run_manifest_artifact:
            step["manifest"] = step_run_manifest_artifact

    return dbt_run_meta_data


def collect_data_from_dbt_cloud(
    dbt_cloud_account_id, dbt_cloud_api_token, dbt_job_run_id
):
    """
    Collect metadata for a single run from dbt cloud.

    Args:
        dbt_cloud_account_id: dbt cloud account id in order to  identify with dbt cloud API
        dbt_cloud_api_token: Api token in order to authenticate dbt cloud API
        dbt_job_run_id: run id of the dbt run that we want to report it's metadata.

        @task
        def prepare_data():
            collect_data_from_dbt_cloud(
            dbt_cloud_account_id=my_dbt_cloud_account_id,
            dbt_cloud_api_token="my_dbt_cloud_api_token",
            dbt_job_run_id=12345
            )
    """
    if not dbt_job_run_id:
        logger.warning("Can't collect run  Data from dbt cloud,missing run id")
        return

    if not dbt_cloud_api_token or not dbt_cloud_account_id:
        logger.warning(
            "Can't collect Data from dbt cloud, account id nor api key were supplied"
        )
        return

    try:
        dbt_cloud_client = DbtCloudApiClient(
            account_id=dbt_cloud_account_id, dbt_cloud_api_token=dbt_cloud_api_token
        )

        dbt_run_meta_data = get_run_data_from_dbt(dbt_cloud_client, dbt_job_run_id)
        if not dbt_run_meta_data:
            return

        _report_dbt_metadata(dbt_run_meta_data)
    except Exception as e:
        logger.warning(
            "Failed collect and report data from dbt cloud api,continue execution"
        )
        log_exception("Could not collect data from dbt cloud", e)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from collections import OrderedDict
from typing import Any, Dict, List, Optional
from uuid import UUID

import six

from six import iteritems

from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.configuration.config_value import ConfigValuePriority
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.constants import RESULT_PARAM
from dbnd._core.parameter.parameter_builder import ParameterFactory
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _ParameterKind,
)
from dbnd._core.parameter.parameter_value import Parameters
from dbnd._core.task_build.decorated_callable_param_builder import (
    DecoratedCallableParamBuilder,
)
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_build.task_signature import user_friendly_signature
from dbnd._core.task_build.task_source_code import TaskSourceCode
from dbnd._core.utils.basics.nothing import is_defined
from dbnd._core.utils.structures import combine_mappings
from dbnd._core.utils.uid_utils import get_uuid


if typing.TYPE_CHECKING:
    from dbnd._core.task_build.task_decorator import TaskDecorator

logger = logging.getLogger(__name__)


def _ordered_params(x):
    return OrderedDict(sorted(x.items(), key=lambda y: y[1].parameter_id))


class TaskDefinition(object):
    """
    TaskDefinition contains all the information gathered and calculated, previous to the task creation.
    """

    @classmethod
    def from_task_cls(cls, task_class, classdict):
        """
        Creating the task definition for a task defined as a class or orchestration decorated task
        This is getting called from the creation of the class itself (by the meta-class)
        """

        # collecting the definitions of the inherited task classes (if any).
        base_task_definitions = get_base_task_definitions(task_class)

        return TaskDefinition(
            classdict=classdict,
            base_task_definitions=base_task_definitions,
            task_passport=TaskPassport.from_task_cls(task_class),
            defaults=classdict.get("defaults", None),
            task_decorator=task_class.task_decorator,
            source_code=TaskSourceCode.from_task_class(task_class),
        )

    @classmethod
    def from_task_decorator(cls, task_decorator):
        # type: (TaskDecorator) -> TaskDefinition
        """
        Creating the task definition from a decorated function.
        This requires the information collected on the function signature
        """
        return TaskDefinition(
            task_passport=task_decorator.task_passport,
            defaults=task_decorator.task_defaults,
            task_decorator=task_decorator,
            source_code=TaskSourceCode.from_callable(
                task_decorator.original_class_or_func
            ),
        )

    def __init__(
        self,
        task_passport,  # type: TaskPassport
        classdict=None,  # type: Optional[Dict[str, Any]]
        base_task_definitions=None,  # type: Optional[List[TaskDefinition]]
        defaults=None,  # type: Optional[Dict[ParameterDefinition, Any]]
        task_decorator=None,  # type: Optional[TaskDecorator]
        source_code=None,  # type: Optional[TaskSourceCode]
        external_parameters=None,  # type: Optional[Parameters]
        task_definition_uid=None,  # type: Optional[UUID]
    ):
        super(TaskDefinition, self).__init__()

        self.hidden = False

        self.task_passport = task_passport
        self.source_code = source_code
        self.task_decorator = task_decorator
        self.base_task_definitions = (
            base_task_definitions or []
        )  # type: List[ TaskDefinition]

        # TODO: maybe use properties or other way to delegate those...
        self.full_task_family = self.task_passport.full_task_family
        self.full_task_family_short = self.task_passport.full_task_family_short
        self.task_family = self.task_passport.task_family
        self.task_config_section = self.task_passport.task_config_section

        # all the attributes that points to_Parameter
        self.task_param_defs = dict()  # type: Dict[str, ParameterDefinition]

        # the defaults attribute
        self.defaults = dict()  # type: Dict[ParameterDefinition, Any]

        self.task_param_defs = self._calculate_task_class_values(
            classdict, external_parameters
        )
        # if we have output params in function arguments, like   f(some_p=parameter.output)
        # the new function can not return the result of return
        self.single_result_output = self._is_result_single_output(self.task_param_defs)

        self.param_defaults = {
            p.name: p.default
            for p in self.task_param_defs.values()
            if is_defined(p.default)
        }

        # TODO: consider joining with task_config
        # TODO: calculate defaults value as _ConfigStore and merge using standard mechanism
        self.defaults = self._calculate_task_defaults(defaults)
        self.task_defaults_config_store = parse_and_build_config_store(
            source=self.task_passport.format_source_name("task.defaults"),
            config_values=self.defaults,
            priority=ConfigValuePriority.FALLBACK,
        )

        self.task_signature_extra = {}
        if config.getboolean("task_build", "sign_with_full_qualified_name"):
            self.task_signature_extra["full_task_family"] = self.full_task_family
        if config.getboolean("task_build", "sign_with_task_code"):
            self.task_signature_extra["task_code_hash"] = user_friendly_signature(
                self.source_code.task_source_code
            )

        if task_definition_uid:
            self.task_definition_uid = task_definition_uid
        else:
            self.task_definition_uid = get_uuid()

    def _calculate_task_class_values(self, classdict, external_parameters):
        # type: (Optional[Dict],  Optional[Parameters]) -> Dict[str, ParameterDefinition]
        # reflect inherited attributes
        params = dict()
        # params will contain definition of param, even it's was overrided by the parent task
        for base_schema in self.base_task_definitions:
            params = combine_mappings(params, base_schema.task_param_defs)

        # let update params with new class attributes
        self._update_params_from_attributes(classdict, params)

        # this is the place we add parameters from function definition
        if self.task_decorator is not None:
            func_params_builder = DecoratedCallableParamBuilder(
                base_params=params, task_decorator=self.task_decorator
            )

            func_params_builder.build_func_params()
            params_dict = dict(func_params_builder.decorator_kwargs_params)
            params_dict.update(func_params_builder.func_spec_params)
            params_dict.update(func_params_builder.result_params)

            self._update_params_from_attributes(params_dict, params)

        if external_parameters:
            params.update(
                {param.name: param for param in external_parameters.get_params()}
            )

        updated_params = {}
        for name, param in six.iteritems(params):
            # add parameters config
            param_with_owner = param.evolve_with_owner(task_definition=self, name=name)

            # updated the owner in the external parameters
            param_value = external_parameters and external_parameters.get_param_value(
                name
            )
            if param_value:
                param_value.parameter = param_with_owner

            updated_params[name] = param_with_owner

        params = _ordered_params(updated_params)
        return params

    def _calculate_task_defaults(self, defaults):
        # type: (...)->  Dict[str, Any]
        base_defaults = dict()
        for base_schema in self.base_task_definitions:
            base_defaults = combine_mappings(base_defaults, base_schema.defaults)

        return combine_mappings(base_defaults, defaults)

    def _update_params_from_attributes(self, classdict, params):
        class_values = dict()
        if not classdict:
            return

        for a_name, a_obj in iteritems(classdict):
            context = "%s.%s" % (self.task_family, a_name)
            try:
                if isinstance(a_obj, ParameterFactory):
                    params[a_name] = a_obj.build_parameter(context)
                elif isinstance(a_obj, ParameterDefinition):
                    params[a_name] = a_obj
                else:
                    class_values[a_name] = a_obj
            except Exception:
                logger.warning("Failed to process %s" % context)
                raise

        # now, if we have overloads in code ( calculated in task_definition):
        # class T(BaseT):
        #     some_base_t_property = new_value
        for p_name, p_val in iteritems(class_values):
            if p_name not in params:
                continue
            params[p_name] = params[p_name].modify(default=p_val)

    def _is_result_single_output(self, params):
        """
        check that task has only one output and it's output is result
         (there can be sub outputs that are part of result)
        """
        result = params.get(RESULT_PARAM)
        if not result:
            return False
        names = result.names if isinstance(result, FuncResultParameter) else []
        for param_def in self.task_param_defs.values():
            if param_def.system or param_def.kind != _ParameterKind.task_output:
                continue
            if param_def.name in [RESULT_PARAM, "task_band"]:
                continue
            if param_def.name in names:
                continue
            return False
        return True

    def __str__(self):
        return "TaskDefinition(%s)" % self.full_task_family

    def run_name(self):
        if self.task_decorator:
            return "%s()" % self.full_task_family
        return "%s.run()" % self.full_task_family


def get_base_task_definitions(task_class):
    """Accessing the task_definitions of the parents classes of the given `task_class`"""
    task_definitions = []

    for cls in reversed(task_class.__bases__):  # type: TaskMetaclass
        if not hasattr(cls, "task_definition"):
            logger.debug(
                "you should inherit from Task objects only: %s -> %s ", task_class, cls
            )
            continue
        task_definitions.append(cls.task_definition)
    return task_definitions



# © Copyright Databand.ai, an IBM Company 2022
import contextlib
import logging
import os

from typing import Optional

from dbnd._core.configuration.environ_config import (
    dbnd_log_init_msg,
    spark_tracking_enabled,
)
from dbnd._core.tracking.airflow_task_context import AirflowTaskContext
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.seven import import_errors


_IS_SPARK_INSTALLED = None
_SPARK_ENV_FLAG = "SPARK_ENV_LOADED"  # if set, we are in spark

logger = logging.getLogger(__name__)


def _is_dbnd_spark_installed():
    global _IS_SPARK_INSTALLED
    if _IS_SPARK_INSTALLED is not None:
        return _IS_SPARK_INSTALLED
    try:
        try:
            from pyspark import SparkContext  # noqa: F401

            from dbnd_spark import dbnd_spark_bootstrap

            dbnd_spark_bootstrap()
        except import_errors:
            _IS_SPARK_INSTALLED = False
        else:
            _IS_SPARK_INSTALLED = True
    except Exception:
        # safeguard, on any exception
        _IS_SPARK_INSTALLED = False

    return _IS_SPARK_INSTALLED


def verify_spark_pre_conditions():
    if spark_tracking_enabled() and _SPARK_ENV_FLAG in os.environ:
        if _is_dbnd_spark_installed():
            return True
        else:
            dbnd_log_init_msg("No Spark: failed to import pyspark or dbnd-spark")
    else:
        dbnd_log_init_msg(
            "No Spark: DBND__ENABLE__SPARK_CONTEXT_ENV or SPARK_ENV_LOADED are not set"
        )
    return False


def _safe_get_active_spark_context():
    if not verify_spark_pre_conditions():
        return None
    try:
        from pyspark import SparkContext

        if SparkContext._jvm is not None:
            return SparkContext._active_spark_context
        else:
            # spark context is not initialized at this step
            logger.info("SparkContext._jvm is not set")
    except Exception as ex:
        logger.info("Failed to get SparkContext: %s", ex)


def _safe_get_jvm_view():
    try:
        spark_context = _safe_get_active_spark_context()
        if spark_context is not None:
            return spark_context._jvm
    except Exception as ex:
        logger.info("Failed to get jvm from SparkContext : %s", ex)


def _safe_get_spark_conf():
    try:
        spark_context = _safe_get_active_spark_context()
        if spark_context is not None:
            return spark_context.getConf()
    except Exception as ex:
        logger.info("Failed to get SparkConf from SparkContext : %s", ex)


def try_get_airflow_context_from_spark_conf():
    # type: ()-> Optional[AirflowTaskContext]
    try:
        conf = _safe_get_spark_conf()
        if conf is not None:
            dag_id = conf.get("spark.env.AIRFLOW_CTX_DAG_ID")
            execution_date = conf.get("spark.env.AIRFLOW_CTX_EXECUTION_DATE")
            task_id = conf.get("spark.env.AIRFLOW_CTX_TASK_ID")
            try_number = conf.get("spark.env.AIRFLOW_CTX_TRY_NUMBER")
            airflow_instance_uid = conf.get("spark.env.AIRFLOW_CTX_UID")

            if dag_id and task_id and execution_date:
                return AirflowTaskContext(
                    dag_id=dag_id,
                    execution_date=execution_date,
                    task_id=task_id,
                    try_number=try_number,
                    airflow_instance_uid=airflow_instance_uid,
                )
            else:
                logger.warning("Airflow context could not be loaded from spark conf")
    except Exception as ex:
        logger.info("Failed to get airflow context info from spark job: %s", ex)

    return None


def get_value_from_spark_env(key):
    try:
        conf = _safe_get_spark_conf()
        value = conf.get("spark.env." + key)
        if value:
            return value
    except:
        return None


@contextlib.contextmanager
def jvm_context_manager(parent_task, current_task):
    """
    This context manager handles external JVM context.
    When task is started we need to explicitly tell JVM
    about it, so it will report metrics to the proper task.
    When task is exited, we need to get context back to the parent task.
    Consider this example:

        # before this task will start, we instruct JVM that current context is changed
        @task
        def parent_task():
            # before starting nested task need we switch context to the nested
            nested_task()
            # after nested task was completed, we need to switch context back to the parent

        @task
        def nested_task():
            # ...
    """
    try:
        set_jvm_context(current_task)
        yield
    finally:
        set_jvm_context(parent_task)


def set_jvm_context(task_run):
    """
    When pyspark is called on the first place we want to ensure that spark listener will report metrics
    to the proper task. To achieve this, we directly set current context to our JVM wrapper.
    """
    try:
        jvm = _safe_get_jvm_view()
        if jvm is None:
            return
        jvm_dbnd = jvm.ai.databand.DbndWrapper

        from py4j import java_gateway

        if isinstance(jvm_dbnd, java_gateway.JavaPackage):
            # if DbndWrapper class is not loaded then agent or IO listener is not attached
            return
        try:
            jvm_dbnd.instance().setExternalTaskContext(
                str(task_run.run.run_uid),
                str(task_run.task_run_uid),
                str(task_run.task_run_attempt_uid),
                str(task_run.task_af_id),
            )
        except Exception as jvm_ex:
            logger.info(
                "Failed to set DBND context to JVM during DbndWrapper call: %s", jvm_ex
            )
    except Exception as ex:
        logger.info("Failed to set DBND context to JVM: %s", ex)


# Logging
def attach_spark_logger(spark_log_file):
    if environ_enabled("DBND__LOG_SPARK"):
        try:
            log4j, spark_logger = try_get_spark_logger()
            if log4j is None:
                return

            pattern = "[%d] {%c,%C{1}} %p - %m%n"
            file_appender = log4j.FileAppender()

            file_appender.setFile(spark_log_file.path)
            file_appender.setName(spark_log_file.path)
            file_appender.setLayout(log4j.PatternLayout(pattern))
            file_appender.setThreshold(log4j.Priority.toPriority("INFO"))
            file_appender.activateOptions()
            spark_logger.addAppender(file_appender)
        except Exception as task_ex:
            logger.warning(
                "Failed to attach spark logger for log %s: %s", spark_log_file, task_ex
            )


def detach_spark_logger(spark_log_file):
    if environ_enabled("DBND__LOG_SPARK"):
        try:
            log4j, spark_logger = try_get_spark_logger()
            if log4j is None:
                return

            spark_logger.removeAppender(spark_log_file.path)
        except Exception as task_ex:
            logger.warning(
                "Failed to detach spark logger for log %s: %s", spark_log_file, task_ex
            )


def try_get_spark_logger():
    try:
        jvm = _safe_get_jvm_view()
        if jvm is None:
            logger.warning(
                "Spark log is enabled but SparkContext is not available. Consider switching DBND__LOG_SPARK to False."
            )
            return None, None
        log4j = jvm.org.apache.log4j
        return log4j, log4j.Logger.getLogger("org.apache.spark")
    except Exception:
        logger.warning(
            "Failed to retrieve log4j context from Spark Context. Spark logs won't be captured."
        )
        return None, None



from unittest import TestCase

import hjson as json

class TestDefault(TestCase):
    def test_default(self):
        self.assertEqual(
            json.dumpsJSON(type, default=repr),
            json.dumpsJSON(repr(type)))



# © Copyright Databand.ai, an IBM Company 2022

import typing

from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, Optional

from targets.value_meta import ValueMetaConf


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run_tracker import TaskRunTracker


class _TaskCtrlMixin(object):
    @property
    @abstractmethod
    def tracker(self):
        # type: () -> TaskRunTracker
        """
        Abstract property for each task to describe how to access it's tracker.
        """
        raise NotImplementedError()

    def log_dataframe(
        self,
        key,
        df,
        with_preview=True,
        with_schema=True,
        with_size=True,
        with_stats=False,
    ):
        meta_conf = ValueMetaConf(
            log_preview=with_preview,
            log_schema=with_schema,
            log_size=with_size,
            log_stats=with_stats,
        )
        self.tracker.log_data(key, df, meta_conf=meta_conf)

    def log_metric(self, key, value, source=None):
        """
        Logs the passed-in parameter under the current run, creating a run if necessary.
        :param key: Parameter name (string)
        :param value: Parameter value (string)
        """
        return self.tracker.log_metric(key, value, source=source)

    def log_metrics(self, metrics_dict, source=None, timestamp=None):
        # type: (Dict[str, Any], Optional[str], Optional[datetime]) -> None
        """
        Logs the all the metrics in the metrics dict to the task tracker.
        @param metrics_dict: name-value pairs of metrics to log
        @param source: optional name of the metrics source
        @param timestamp: optional timestamp of the metrics
        """
        return self.tracker.log_metrics(
            metrics_dict, source=source, timestamp=timestamp
        )

    def log_system_metric(self, key, value):
        """Shortcut for log_metric(..., source="system")"""
        return self.log_metric(key, value, source="system")

    def log_artifact(self, name, artifact):
        """Log a local file or directory as an artifact of the currently active run."""
        return self.tracker.log_artifact(name, artifact)

    @property
    @abstractmethod
    def task_dag(self):
        """
        Abstract property for each task to describe how to access it's task dag controller.
        """
        raise NotImplementedError()

    def set_upstream(self, task_or_task_list):
        self.task_dag.set_upstream(task_or_task_list)

    def set_downstream(self, task_or_task_list):
        self.task_dag.set_downstream(task_or_task_list)

    def __lshift__(self, other):
        return self.set_upstream(other)

    def __rshift__(self, other):
        return self.set_downstream(other)

    def set_global_upstream(self, task_or_task_list):
        self.task_dag.set_global_upstream(task_or_task_list)

    @property
    @abstractmethod
    def descendants(self):
        raise NotImplementedError()



# MIT LICENCE
# from https://github.com/skorokithakis/tbvaccine/commit/82529b2b91d38d20f1814cfa6701b2169b888797

# BASED ON https://github.com/skorokithakis/tbvaccine/
# Original code is licenced with

# The MIT License (MIT)
#
# Copyright (c) 2016 Stavros Korokithakis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
import sys
import traceback

from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter as TerminalFormatter
from pygments.lexers.python import PythonLexer

#  term colour control codes
from dbnd._core.errors.errors_utils import UserCodeDetector


re_ansi_control_codes = re.compile(r"\x1b[^m]*m")


class State:
    no_idea = 0
    in_traceback = 1


class TBVaccine:
    def __init__(
        self,
        user_code_detector=None,
        isolate=True,
        show_vars=False,
        max_length=120,
        skip_non_user_on_isolate=True,
        no_colors=False,
        color_scheme="monokai",
    ):
        self._user_code_detector = (
            user_code_detector or UserCodeDetector.build_code_detector()
        )

        self.skip_non_user_on_isolate = skip_non_user_on_isolate and isolate
        # Whether to print interesting lines in color or not. If False,
        # all lines are printed in color.
        self._isolate = isolate
        self._no_colors = no_colors

        # Whether to print variables for stack frames.
        self._show_vars = show_vars and isolate

        # Max length of printed variable lines
        self._max_length = max_length

        self.color_scheme = color_scheme

        try:
            self.pygments_lexer = PythonLexer()
            self.pygments_formatter = TerminalFormatter(style=self.color_scheme)
        except Exception:
            self.pygments_lexer = None
            self.pygments_formatter = None

    def _new_tb_print(self):
        return _TBMessage(self)

    def _highlight_line(self, line):
        if self._no_colors or not self.pygments_lexer:
            return line.rstrip("\r\n")

        line = highlight(line, self.pygments_lexer, self.pygments_formatter)
        return line.rstrip("\r\n")

    def _highlight_text(self, text, fg=None, style=None, max_length=None):
        if self._no_colors:
            return text

        raw_text = re.sub(re_ansi_control_codes, "", text)
        if max_length and len(raw_text) > max_length:
            short_text = text[: int(max_length * 3)]
            # Check if there's an ANSI escape in the last few chars of max_length and break before it.
            if "\x1b" in short_text[-10:]:
                short_text = short_text[: short_text.rfind("\x1b")]
            text = short_text + "\x1b[0m ... ({} more chars)".format(
                len(text) - len(short_text)
            )
        if fg or style:
            styles = {"bright": 1, None: 0}
            colors = {
                "black": 30,
                "red": 31,
                "green": 32,
                "yellow": 33,
                "blue": 34,
                "magenta": 35,
                "cyan": 36,
                "gray": 37,
            }
            text = "\x1b[%d;%dm%s\x1b[m" % (styles[style], colors[fg], text)
        return text

    def format_tb(self, etype, value, tb):
        """
        Format an entire traceback string with ANSI colors.
        """
        return self._new_tb_print().format_tb(etype, value, tb)

    def format_exc(self):
        """
        Format the latest exception's traceback.
        """
        return self.format_tb(sys.exc_info()[2])

    def print_exc(self):
        """
        Format the latest exception's traceback.
        """
        msg = self.format_exc()
        sys.stderr.write(msg)


class _TBMessage(object):
    TB_END_RE = re.compile(r"^(?P<exception>[\w\.]+)\: (?P<description>.*?)$")
    TB_FILE_RE = re.compile(
        r'^  File "(?P<filename>.*?)", line (?P<line>\d+), in (?P<func>.*)$'
    )
    VAR_PREFIX = "|     "

    def __init__(self, tbv):
        # type: (_TBMessage, TBVaccine) -> None
        self.tbv = tbv
        # Our current state.
        self._state = State.no_idea

        # The filename of the line we're currently printing.
        self._file = None

        # The buffer that we use to build up the output in.
        self._buffer = ""

        self._found_user_code = False
        self._current_filtered_lines = []

    def _user_space(self):
        """
        Decide whether the file in the traceback is one in our code_dir or not.
        """
        if not self.tbv._user_code_detector:
            return True
        return self.tbv._user_code_detector.is_user_file(self._file)

    def _print(self, text, fg=None, style=None, max_length=None):

        self._buffer += self.tbv._highlight_text(
            text, fg=fg, style=style, max_length=max_length
        )

    def _process_var_line(self, line):
        """
        Process a line of variables in the traceback.
        """
        if not self.tbv._show_vars:
            # Don't print.
            return False

        if not self._user_space():
            return False

        line = self.tbv._highlight_line(line)
        self._print(line, max_length=self.tbv._max_length)
        return True

    def _process_code_line(self, line):
        """
        Process a line of code in the traceback.
        """
        if not self._user_space():
            if self.tbv.skip_non_user_on_isolate and not self._found_user_code:
                return False
            # Print without colors.
            self._print(line)
        else:
            if self.tbv._isolate:
                line = line[1:]
                self._print(">", fg="red", style="bright")
            self._print(line, fg="gray")

        return True

    def _process_file_line(self, line):
        """
        Process a "file" line of traceback.
        """
        match = self.TB_FILE_RE.match(line).groupdict()
        self._file = match["filename"]

        if not self._user_space():
            if self.tbv.skip_non_user_on_isolate and not self._found_user_code:
                return False
            # Print without colors.
            self._print(line)
        else:
            # let mark that we have seen some user code
            self._found_user_code = True
            self._print('  File "')
            base, fn = os.path.split(match["filename"])
            if base:
                self._print(base + os.sep, "cyan")
            self._print(fn, "cyan", style="bright")
            self._print('", line ')
            self._print(match["line"], "yellow")
            self._print(", in ")
            self._print(match["func"], "magenta")
        return True

    def _process(self, lines):
        """
        Process a line of input.
        """
        for line in lines:
            sl = line.rstrip("\r\n")
            if not sl:
                continue

            printed = True
            if self._state == State.no_idea:
                if sl == "Traceback (most recent call last):":
                    self._found_user_code = False
                    self._current_filtered_lines = []
                    # The first line of the traceback.
                    self._state = State.in_traceback
                    self._print(sl, "blue")
                else:
                    self._print(sl)
            elif self._state == State.in_traceback:
                if self.TB_END_RE.match(sl):
                    # The last line of the traceback.
                    if not self._found_user_code and self._current_filtered_lines:
                        self._print("\n".join(self._current_filtered_lines))
                    self._state = State.no_idea
                    matches = self.TB_END_RE.match(sl).groupdict()
                    self._print(matches["exception"], "red", "bright")
                    self._print(": ")
                    self._print(matches["description"], "green")
                    self._print("\n")
                elif self.TB_FILE_RE.match(line):
                    # A file line.
                    printed = self._process_file_line(sl)
                elif sl.startswith(self.VAR_PREFIX):
                    printed = self._process_var_line(sl)
                elif sl.startswith("    "):
                    # A code line.
                    printed = self._process_code_line(sl)
                else:
                    self._print(sl)

            if printed:
                self._print("\n")
            else:
                self._current_filtered_lines.append(line)

    # def format_tb_with_locals(self, etype, value, tb):
    #     """
    #     Return a traceback as a string, with the local variables in each stack.
    #     """
    #     original_tb = tb
    #     while True:
    #         if not tb.tb_next:
    #             break
    #         tb = tb.tb_next
    #
    #     stack = []
    #     f = tb.tb_frame
    #     while f:
    #         stack.append(f)
    #         f = f.f_back
    #     stack.reverse()
    #
    #     lines = ["Traceback (most recent call last):\n"]
    #     for frame, line in zip(stack, traceback.format_tb(original_tb)):
    #         # Frame lines contain newlines, so we need to split on them.
    #         lines.extend(line.split("\n"))
    #         var_tuples = sorted(frame.f_locals.items())
    #         if not var_tuples:
    #             # There are no locals, so continue.
    #             continue
    #
    #         max_length = max([len(x[0]) for x in var_tuples])
    #         for key, val in var_tuples:
    #             if type(val) in (type(lambda: None), type(sys)) or (
    #                 key.startswith("__") and key.endswith("__")
    #             ):
    #                 # We don't want to print functions or modules or __variables__.
    #                 continue
    #             try:
    #                 val = str(val)
    #             except:  # noqa
    #                 val = "<CANNOT CONVERT VALUE>"
    #             lines.append(
    #                 "%s%s = %s" % (self.VAR_PREFIX, key.ljust(max_length), val)
    #             )
    #     lines.append("%s: %s" % (value.__class__.__name__, value))
    #
    #     for line in lines:
    #         self._process_line(line)
    #
    #     return self._buffer

    def format_tb(self, etype, value, tb):
        """
        Format an entire traceback string with ANSI colors.
        """

        tb_string = traceback.format_exception(etype, value, tb)
        tb_string = "".join(tb_string)
        self._process(tb_string.split("\n"))
        return self._buffer


def add_hook(*args, **kwargs):
    if not getattr(sys.stderr, "isatty", lambda: False)():
        return
    tbv = TBVaccine(*args, **kwargs)
    sys.excepthook = tbv.print_exc



# © Copyright Databand.ai, an IBM Company 2022

from typing import List

from targets.extras import DataTargetCtrl
from targets.marshalling import get_marshaller_ctrl
from targets.target_config import file


class ObjectMarshallingCtrl(DataTargetCtrl):
    def _load(self, value_type, config, **kwargs):
        marshaler_ctrl = get_marshaller_ctrl(
            target=self.target, value_type=value_type, config=config
        )
        return marshaler_ctrl.load(**kwargs)

    def _dump(self, value_type, config, value, **kwargs):
        marshaler_ctrl = get_marshaller_ctrl(
            target=self.target, value_type=value_type, config=config
        )
        return marshaler_ctrl.dump(value, **kwargs)

    # aliases fro the most used functions - do things regardless file type
    def readlines(self, **kwargs):
        return self._load(List[str], file.txt, **kwargs)

    def read(self, **kwargs):
        return self._load(object, file.txt, **kwargs)

    def read_obj(self, **kwargs):
        return self._load(object, file.pickle, **kwargs)

    def read_json(self, **kwargs):
        return self._load(object, file.json, **kwargs)

    def read_pickle(self, **kwargs):
        return self._load(object, file.pickle, **kwargs)

    def write(self, fn, **kwargs):
        self._dump(object, file.txt, fn, **kwargs)

    def writelines(self, fn, **kwargs):
        self._dump(List[str], file.txt, fn, **kwargs)

    def write_pickle(self, obj, **kwargs):
        self._dump(object, file.pickle, obj, **kwargs)

    def write_numpy_array(self, arr, **kwargs):
        import numpy

        self._dump(numpy.ndarray, file.numpy, arr, **kwargs)

    def write_json(self, obj, **kwargs):
        self._dump(object, file.json, obj, **kwargs)

    def touch(self):
        with self.target.open("w"):
            pass



# © Copyright Databand.ai, an IBM Company 2022

import functools

from collections.abc import Mapping

import six

from dbnd._core.utils.type_check_utils import is_instance_by_class_name


class DatabandDict(object):
    pass


def flatten(struct):
    """
    Creates a flat list of all all items in structured output (dicts, lists, items):

    .. code-block:: python

        >>> sorted(flatten({'a': 'foo', 'b': 'bar'}))
        ['bar', 'foo']
        >>> sorted(flatten(['foo', ['bar', 'troll']]))
        ['bar', 'foo', 'troll']
        >>> flatten('foo')
        ['foo']
        >>> flatten(42)
        [42]
    """
    if struct is None:
        return []
    flat = []
    if isinstance(struct, (dict, Mapping)):
        for _, result in six.iteritems(struct):
            flat += flatten(result)
        return flat
    if isinstance(struct, six.string_types):
        return [struct]

    if getattr(struct, "target_no_traverse", None):
        return [struct]

    try:
        # if iterable
        iterator = iter(struct)
    except TypeError:
        return [struct]

    for result in iterator:
        flat += flatten(result)
    return flat


def f_identity(x):
    return x


def f_not_none(x):
    return x


def f_no_filter(x):
    return True


def traverse(
    struct,
    convert_f=f_identity,
    filter_none=False,
    filter_empty=False,
    convert_types=None,
    traverse_path=None,
):
    """
    Maps all Tasks in a structured data object to their .output().
    :param convert_types non basic types to apply convert_f
    :param filter_none  remove None from structures
    """
    obj = struct

    if traverse_path is None:
        traverse_path = []  # first invocation fo the function

    t = functools.partial(
        traverse,
        convert_f=convert_f,
        filter_none=filter_none,
        filter_empty=filter_empty,
        convert_types=convert_types,
    )

    if convert_types and isinstance(obj, convert_types):
        return convert_f(obj)

    if isinstance(obj, Mapping):
        # noinspection PyArgumentList
        converted = (
            (k, t(v, traverse_path=traverse_path + [k])) for k, v in six.iteritems(obj)
        )
        if filter_none:
            converted = ((k, v) for k, v in converted if v is not None)
        new_obj = obj.__class__(converted)
        if filter_empty and not new_obj:
            return None
        return new_obj

    if isinstance(obj, six.string_types):
        return convert_f(obj)

    if "pyspark" in str(type(obj)):
        return convert_f(obj)

    try:
        target_no_traverse = hasattr(obj, "target_no_traverse") and getattr(
            obj, "target_no_traverse", None
        )
    except ValueError:
        # SPARK OBJECTS do not support hasattr
        target_no_traverse = True
    if target_no_traverse is bool and target_no_traverse:
        return convert_f(obj)

    list_obj_constructor = None
    if isinstance(obj, (list, tuple, set)):
        list_obj_constructor = obj.__class__
    elif is_instance_by_class_name(obj, "DataFrame"):
        pass
    else:
        try:
            iter(obj)  # noqa: F841
            list_obj_constructor = list
        except TypeError:
            pass

    # we can parse and reconstruct list object
    if list_obj_constructor:
        converted = (t(r, traverse_path=traverse_path + [i]) for i, r in enumerate(obj))
        if filter_none:
            converted = filter(lambda x: x is not None, converted)
        new_obj = list_obj_constructor(converted)
        if filter_empty and not new_obj:
            return None
        return new_obj

    # so it's simple obj, let apply function
    return convert_f(obj)


def _frozen_set(struct):
    if isinstance(struct, set):
        from dbnd._core.utils import json_utils

        return sorted(struct, key=lambda x: json_utils.dumps_canonical(x))
    return struct


def getpaths(struct):
    """
    Maps all Tasks in a structured data object to their .output().
    """
    return traverse(struct, lambda t: t.task_outputs)


def traverse_to_str(obj):
    if obj is None:
        return None
    return traverse(obj, convert_f=str)


def traverse_frozen_set(obj):
    return traverse(obj, convert_types=(set,), convert_f=_frozen_set)



# © Copyright Databand.ai, an IBM Company 2022

import os


def abs_join(*path):
    return os.path.abspath(os.path.normpath(os.path.join(*path)))


def relative_path(file_in_path, *path):
    """
    Returns absolute path of a local file/directory.

    Args:
        file_in_path: Local file.
        *path: The path you want to turn into an absolute path.

    Returns: Absolute path of *path.
    """
    current_dir = os.path.dirname(os.path.abspath(file_in_path))
    return abs_join(current_dir, *path)


def relative_path_directory(file_in_path, *path):
    current_dir = os.path.abspath(file_in_path)
    return abs_join(current_dir, *path)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor.marshmallow import fields, pre_load


class MLAlert(ApiStrictSchema):
    sensitivity = fields.Float()
    look_back = fields.Integer()


class AlertDefsSchema(ApiStrictSchema):
    severity = fields.Str(required=True)
    type = fields.Str(required=True)
    user_metric = fields.Str()
    operator = fields.Str()
    is_str_value = fields.Bool()

    created_at = fields.DateTime()
    # TODO_CORE: API: Deprecate airflow_server_info
    project_id = fields.Int(attribute="job.project_id")
    project_name = fields.Str(attribute="job.project.name")
    group_uid = fields.Str(allow_none=True, load_from="alert_group_uid")
    root_group_uid = fields.Str(allow_none=True, load_from="alert_root_group_uid")

    uid = fields.Str(allow_none=True)
    value = fields.Str(allow_none=True)
    job_id = fields.Int(allow_none=True)
    job_name = fields.Str(attribute="job.name", allow_none=True)
    task_repr = fields.Str(allow_none=True)
    task_name = fields.Str(allow_none=True)
    custom_name = fields.Str(allow_none=True)
    original_uid = fields.Str(allow_none=True)
    advanced_json = fields.Str(allow_none=True)
    scheduled_job_uid = fields.Str(allow_none=True)
    scheduled_job_name = fields.Str(attribute="scheduled_job.name", allow_none=True)
    custom_description = fields.Str(allow_none=True)
    ml_alert = fields.Nested(MLAlert, allow_none=True)

    # Fields for DatasetSlaAlert/DatasetSlaAdvancedAlert alert
    # --------------------------------------
    seconds_delta = fields.Int(allow_none=True)  # Converts to datetime.timedelta
    dataset_partial_name = fields.Str(allow_none=True)
    datasets_uids = fields.List(fields.Str(), allow_none=True)

    # Fields for OperationColumnStatAdvancedAlert alert
    # --------------------------------------
    dataset_uid = fields.Str(allow_none=True)
    # Operation type (e.g. "read", "write", None=any) to filter stats by
    operation_type = fields.Str(allow_none=True)

    # Type of MetricRule, found in dbnd_web. Used to build advanced_json
    metrics_rules = fields.List(fields.Dict(), allow_none=True)

    @pre_load
    def prepere(self, data: dict, **kwargs):
        value = data.get("value", None)
        if value is not None:
            data["value"] = str(data["value"])
        return data



# © Copyright Databand.ai, an IBM Company 2022

import logging


logger = logging.getLogger(__name__)


def start_pycharm_debugger(host, port):
    """
    This is starting a debug connection to the pycharm at the address '{host}:{port}'
    """
    logger.info(
        "Starting debugging session with `pydevd_pycharm` on '{host}:{port}'".format(
            host=host, port=port
        )
    )

    try:
        import pydevd_pycharm
    except ImportError:
        logger.warning(
            "\n\tCouldn't import `pydevd_pycharm`.\n"
            "\tHelp:\tMake sure to install the relevant `pydevd_pycharm` version matching to your pycharm version"
        )
        return

    try:
        pydevd_pycharm.settrace(
            host, port=port, stdoutToServer=True, stderrToServer=True
        )
        return
    except ConnectionError:
        logger.exception(
            "\n\tCouldn't connect to pycharm host at '{host}:{port}'.\n"
            "\tHelp:\tPlease assert that there is a Python Debug Server listening on that address.\n"
            "\tFor more info:\thttps://www.jetbrains.com/help/pycharm/remote-debugging-with-product.html#remote-debug-config".format(
                host=host, port=port
            )
        )
        return



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.constants import TaskType
from dbnd._core.errors.friendly_error.task_execution import (
    failed_to_assign_result,
    failed_to_process_non_empty_result,
)
from dbnd._core.task.pipeline_task import PipelineTask
from dbnd._core.task.python_task import PythonTask
from dbnd._core.task.task import Task
from dbnd._core.task_build.task_results import FuncResultParameter


logger = logging.getLogger(__name__)


class _DecoratedCallableTask(Task):
    _dbnd_decorated_task = True

    # default result for any decorated function
    result = None

    def _invoke_func(self, extra_kwargs=None, force_invoke=False):
        # this function is in charge of calling user defined code (decorated function) call
        # usually it's called from from task.run/task.band
        extra_kwargs = extra_kwargs or {}
        spec = self.task_decorator.get_callable_spec()
        invoke_kwargs = {}
        for name in spec.args:
            # if there is no parameter - it was disabled at TaskDefinition building stage
            if self._params.get_param(name) is None:
                continue
            invoke_kwargs[name] = getattr(self, name)

        invoke_kwargs.update(extra_kwargs)

        if not self._dbnd_call_state:
            from dbnd._core.task_build.task_cls__call_state import TaskCallState

            self._dbnd_call_state = TaskCallState()
        self._dbnd_call_state.start()

        if self.task_decorator.is_class:
            # this is the case of
            # @task
            # class UserClass:
            #     pass
            # now we are in the Task instance, it was created via UserClass() at @pipeline
            obj_cls = self.task_decorator.class_or_func
            invoke_kwargs["__call_original_cls"] = True
            self.task_user_obj = obj_cls(**invoke_kwargs)
            result = self.task_user_obj.run()
        else:
            # we are going to run user function
            func_call = spec.item
            result = func_call(**invoke_kwargs)

        self._dbnd_call_state.finish(result)

        result_param = self.__class__.result
        if result_param is None and result:
            raise failed_to_process_non_empty_result(self, result)

        if isinstance(result_param, FuncResultParameter):
            # if we have result that combined from different output params
            # assign all returned values to relevant outputs
            # so they will be automatically saved
            if result is None:
                raise failed_to_assign_result(self, result_param)
            for r_name, value in result_param.named_results(result):
                setattr(self, r_name, value)
        else:
            self.result = result
        return result

    def on_kill(self):
        task_user_obj = getattr(self, "task_user_obj", None)
        if task_user_obj is not None and hasattr(task_user_obj, "on_kill"):
            task_user_obj.on_kill()
            return
        else:
            super(_DecoratedCallableTask, self).on_kill()


class DecoratedPythonTask(PythonTask, _DecoratedCallableTask):
    _conf__task_type_name = TaskType.python

    def run(self):
        self._invoke_func(force_invoke=True)


class DecoratedPipelineTask(PipelineTask, _DecoratedCallableTask):
    _conf__task_type_name = TaskType.pipeline

    def band(self):
        return self._invoke_func()



# © Copyright Databand.ai, an IBM Company 2022

"""
Task build process
=================
TaskFactory is the class that is responsible for Task/Config object creation

Class.param is calculated with following rules (top to low priority):
 1. Config value with Override priority
 2. ctor(param=VALUE)             # regular ctor
 3. Config value
 4. Parent task parameter with the same name and defined as paramter(scope=ParameterScope.children)
 5. if paramter(task_from_env=True), VALUE := Task.task_env.parameter
 6. parameter(default=VALUE)

We calculate task_env and task_config first as other parameters can be change based on their value
Following config layers are added on top of existing config of configuration, environment,
 cli and other parent tasks:
     Task.defaults  with priority FALLBACK
     Task.task_config or ctor(task_config=..)
     override_config from ctor(override=override_config) with priority OVERRIDE

Config Value are calculated based on following sections (top to low priority)
  * task_name
  * task_family
  * full_task_family
  * _from=SECTION pragma at configuration
  * task for Task, config for Config

TLDR:
    Task.defaults
    -------------

    Task.defaults is the lowest level of config (DEFAULTS)
     * It's a property of the Task class. All values are merged when TaskB inherits from TaskA (at TaskDefinition)
     * task.defaults value is calculated at TaskDefinition during the "compilation of the Task class definition"
    Task.defaults value is merged into Current config (with FALLBACK priority)
     * do not support override()!!!
     * there is no "merge" behavior for values. (Dictionaries are going to be replaced)
         class TaskA(...):
            defaults = {SparkConf.conf : { 1:1} , SparkConf.jar : "taskajar }

         class TaskB(TaskA):
            defaults = {SparkConf.conf : { 1:2}  }
        TaskB.defaults == {SparkConf.conf : { 1:2} , SparkConf.jar : "taskajar }

    Task.task_config
    ----------------

    Task.task_config - extra config level
     * it "replaces" inherited task_config from super class if exists
     * it is calculated every time we create Task
     * supports override():  `task_config = {SparkConf.jar: override("my_jar")}`
     * Usually, it's used with some basic "nested" config param: like SparkConfig.conf.
     Task.task_config issues:
       1.     task_config = {SparkConf.jar: "my_jar"}
             [spark_remote]
             jar=from_config
        User will get my_jar when engine is [spark] and  "from_config" if engine is [spark_remote].
        The reason is that engine looks for the value in most relevant section which is [spark_remote],
        while SparkConf.jar provides a value for [spark] section
        The only way to workaround it right now is to use:
            task_config = {SparkConf.jar: override("my_jar")}
        The system will look for override values and if it exists,
        the override will be taken (despite best section match).

    Task(overrides=conf_value...)
    -----------------------------
     * uses the same system as task_config, but override is applied to all values


    param = parameter(scope=ParameterScope.children)
    ------------------------------------------------

    Task Parameters Child Scoping - if task defined at
    Task can get the value for some params from Parent task (parent of any level)

    for example:
      class ChildTask(Task):
          my_p = parameter() # it doesn't have to be scoped
      class ParentTask(Task):
          my_p = parameter(scope=ParameterScope.children)
        def band:
          my_p_task = ChildTask() # no explicit "wiring" of my_p



    param = parameter(task_from_env=True)
    ------------------------------------------------

    Task Parameters value default will be taken from Task.task_env[EnvConfig] object


"""
import functools
import logging
import typing

import six

from more_itertools import last, unique_everseen
from six import iteritems

from dbnd._core.configuration.config_path import CONF_CONFIG_SECTION, CONF_TASK_SECTION
from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.configuration.config_store import _ConfigStore
from dbnd._core.configuration.config_value import ConfigValue, ConfigValuePriority
from dbnd._core.configuration.pprint_config import pformat_current_config
from dbnd._core.constants import RESULT_PARAM, ParamValidation, _TaskParamContainer
from dbnd._core.current import get_databand_context
from dbnd._core.errors import MissingParameterError, friendly_error
from dbnd._core.parameter.constants import ParameterScope
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    build_parameter_value,
)
from dbnd._core.parameter.parameter_value import (
    ParameterFilters,
    Parameters,
    ParameterValue,
    fold_parameter_value,
)
from dbnd._core.task_build.task_context import (
    TaskContextPhase,
    task_context,
    try_get_current_task,
)
from dbnd._core.task_build.task_passport import format_source_suffix
from dbnd._core.task_build.task_signature import (
    TASK_ID_INVALID_CHAR_REGEX,
    build_signature,
)
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.basics.text_banner import safe_string
from dbnd._core.utils.callable_spec import args_to_kwargs
from targets import target
from targets.target_config import parse_target_config


if typing.TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Type

    from dbnd._core.configuration.dbnd_config import DbndConfig
    from dbnd._core.settings import EnvConfig
    from dbnd._core.task.task_with_params import _TaskWithParams
    from dbnd._core.task_build.task_definition import TaskDefinition

TASK_BAND_PARAMETER_NAME = "task_band"
logger = logging.getLogger(__name__)


class TaskFactoryConfig(object):
    """(Advanced) Databand's core task builder"""

    config_section = "task_build"

    def __init__(self, verbose, sign_with_full_qualified_name, sign_with_task_code):
        self.verbose = verbose
        self.sign_with_full_qualified_name = sign_with_full_qualified_name
        self.sign_with_task_code = sign_with_task_code

    @classmethod
    def from_dbnd_config(cls, conf):
        # we can not use standard "Config" class creation here, as we are used for that

        def _b(param_name):
            return conf.getboolean(cls.config_section, param_name)

        return cls(
            verbose=_b("verbose"),
            sign_with_full_qualified_name=_b("sign_with_full_qualified_name"),
            sign_with_task_code=_b("sign_with_task_code"),
        )


def get_task_from_sections(config, task_name):
    extra_sections = []
    while task_name:
        # "pseudo" recursive call
        # we check if we have something  for current task_name, and if we do - that's a from
        task_from = config.get(task_name, "_from", None)
        if task_from == task_name:
            # let not throw exception, as it's not a critical error.
            task_from = None

        if task_from:
            extra_sections.append(task_from)
        task_name = task_from

    return extra_sections


class TaskFactory(object):
    """
    Create a TaskWithParam object by
    1. Calculating its params
    2. Register to cache
    3. Calling Tasks initializing methods - _initialize and _validate
    """

    def __init__(self, config, task_cls, task_definition, task_args, task_kwargs):
        # type:(DbndConfig, Type[_TaskWithParams],TaskDefinition, Any, Any)->None
        self.task_cls = task_cls
        self.task_definition = task_definition

        # keep copy of user inputs
        self.task_kwargs__ctor = task_kwargs.copy()
        self.task_args__ctor = list(task_args)

        self.task_env_config = None  # type: Optional[EnvConfig]

        self.parent_task = try_get_current_task()
        self._ctor_as_str = "%s@%s" % (
            _get_call_repr(
                self.task_passport.task_family,
                self.task_args__ctor,
                self.task_kwargs__ctor,
            ),
            str(self.task_cls),
        )

        # extract all "system" keywords from kwargs
        # support task_family in kwargs -> use it as task_name (old behavior)
        task_family = task_kwargs.get("task_family", self.task_passport.task_family)
        self.task_name = task_kwargs.pop("task_name", task_family)
        self.task_config_override = task_kwargs.pop("override", None) or {}
        task_config_sections_extra = task_kwargs.pop("task_config_sections", None)
        self.task_kwargs = task_kwargs

        self.task_name = TASK_ID_INVALID_CHAR_REGEX.sub("_", self.task_name)

        self.task_factory_config = TaskFactoryConfig.from_dbnd_config(config)
        self.verbose_build = self.task_factory_config.verbose

        # current config, NOTE: it's Singleton
        self.config = config
        self.config_sections = []

        self.task_errors = []
        self.build_warnings = []

        # will be used for ConfigValue
        self.config_sections = self._get_task_config_sections(
            config=config, task_config_sections_extra=task_config_sections_extra
        )

    @property
    def task_passport(self):
        return self.task_definition.task_passport

    def _get_config_value_stack_for_param(self, param_def):
        # type: (ParameterDefinition) -> List[Optional[ConfigValue]]
        """
        Wrap the extracting the value of param_def from configuration or config_path
        """
        try:
            config_values_stack = self.config.get_multisection_config_value(
                self.config_sections, key=param_def.name
            )
            if config_values_stack:
                return config_values_stack

            if param_def.config_path:
                config_path = self.config.get_config_value(
                    section=param_def.config_path.section, key=param_def.config_path.key
                )
                if config_path:
                    return [config_path]
            return []
        except Exception as ex:
            raise param_def.parameter_exception("read configuration value", ex)

    def _build_parameter_value(self, param_def):
        # type: (ParameterDefinition) -> ParameterValue
        """
               -= MAIN FUNCTION for Parameter Calculation =-
        This is the place we calculate param_def value
        based on Class(defaults, constructor, overrides, root)
        and Config(env, cmd line, config)  state

        Build parameter value from config_value and param_definition.
        Considerate - priority, constructor values, param type and so on.
        """
        config_values_stack = self._get_config_value_stack_for_param(param_def)
        param_name = param_def.name

        # if we will use config values we will need to reduce them to a single value
        # if there is only one - reduce will get it
        # other wise we need combine the values using `fold_parameter_value`
        # may raise if can't fold two values together!!

        has_overrides_in_config = any(
            cf
            for cf in config_values_stack
            if cf and cf.priority >= ConfigValuePriority.OVERRIDE
        )

        # first check for override (HIGHEST PRIORITY)
        if has_overrides_in_config:
            config_values_as_param_values = [
                build_parameter_value(param_def, cf) for cf in config_values_stack
            ]
            return functools.reduce(fold_parameter_value, config_values_as_param_values)

        # second using kwargs we received from the user
        # do we need to do it for tracking?
        elif param_name in self.task_kwargs:
            return build_parameter_value(
                param_def,
                ConfigValue(
                    self.task_kwargs.get(param_name), source=self._source_name("ctor")
                ),
            )

        if config_values_stack:
            config_values_as_param_values = [
                build_parameter_value(param_def, cf) for cf in config_values_stack
            ]
            return functools.reduce(fold_parameter_value, config_values_as_param_values)

        if (
            self.parent_task
            and param_name in self.parent_task.task_children_scope_params
        ):
            # we have parent task with param = parameter(scope=ParameterScope.children)
            # the priority is lower than config (as this one is more like "default" than "explicit config")
            # see _calculate_task_children_scope_params implementation and ParameterScope.children
            parameter_value = self.parent_task.task_children_scope_params[param_name]
            return build_parameter_value(
                param_def,
                ConfigValue(value=parameter_value.value, source=parameter_value.source),
            )

        if param_def.from_task_env_config:
            # param = parameter(from_task_env_config=True)
            # we check task.task_env.param for the value
            if not self.task_env_config:
                raise friendly_error.task_parameters.task_env_param_with_no_env(
                    context=self._ctor_as_str, key=param_name
                )
            param_env_config_value = self.task_env_config.task_params.get_param_value(
                param_name
            )
            if param_env_config_value is None:
                raise friendly_error.task_parameters.task_env_param_not_exists_in_env(
                    context=self._ctor_as_str,
                    key=param_name,
                    env_config=self.task_env_config,
                )
            return build_parameter_value(
                param_def,
                ConfigValue(
                    value=param_env_config_value.value,
                    source=param_env_config_value.source,
                ),
            )

        if param_name in self.task_definition.param_defaults:
            # we can't "add" defaults to the current config and rely on configuration system
            # we don't know what section name to use, and it my clash with current config
            return build_parameter_value(
                param_def,
                ConfigValue(
                    self.task_definition.param_defaults.get(param_name),
                    source=self._source_name("default"),
                ),
            )

        if not param_def.is_output():
            # outputs can be none, we "generate" their values later
            err_msg = "No value defined for '{name}' at {context_str}".format(
                name=param_name, context_str=self._ctor_as_str
            )
            raise MissingParameterError(
                err_msg,
                help_msg=param_def._get_help_message(sections=self.config_sections),
            )

        # returning empty Output value
        return ParameterValue(
            parameter=param_def,
            source="",
            source_value=NOTHING,
            value=NOTHING,
            parsed=False,
        )

    # should refactored out
    def _get_task_config_sections(self, config, task_config_sections_extra):
        # [HIGHEST PRIORITY, ... , LOWEST PRIORITY]
        sections = [self.task_name]

        # '_from=...' at config files
        sections.extend(get_task_from_sections(config, self.task_name))

        # sections by family
        sections.extend(
            [self.task_passport.task_family, self.task_passport.full_task_family]
        )

        # from user
        if task_config_sections_extra:
            sections.extend(task_config_sections_extra)

        from dbnd._core.task.config import Config

        # adding "default sections"  - LOWEST PRIORITY
        if issubclass(self.task_cls, _TaskParamContainer):
            sections += [CONF_TASK_SECTION]
        elif issubclass(self.task_cls, Config):
            sections += [CONF_CONFIG_SECTION]

        # dedup values and preserve order
        sections = map(lambda x: x.lower(), filter(None, sections))
        sections = list(unique_everseen(sections))

        return sections

    def _get_override_params_signature(self):
        override_signature = {}
        for p_obj, p_val in six.iteritems(self.task_config_override):
            if isinstance(p_obj, ParameterDefinition):
                override_key = "%s.%s" % (p_obj.task_definition.task_family, p_obj.name)
                override_value = (
                    p_val
                    if isinstance(p_val, six.string_types)
                    else p_obj.signature(p_val)
                )
            else:
                # very problematic approach till we fix the override structure
                override_key = str(p_obj)
                override_value = str(p_val)
            override_signature[override_key] = override_value
        return override_signature

    def build_task_object(self, task_metaclass):
        databand_context = get_databand_context()

        # convert args to kwargs, validate values
        self.task_kwargs = self._build_and_validate_task_ctor_kwargs(
            self.task_args__ctor, self.task_kwargs
        )

        self._log_build_step("Resolving task params with %s" % self.config_sections)
        try:
            task_param_values = self._build_task_param_values()
            task_params = Parameters(
                source=self._ctor_as_str, param_values=task_param_values
            )

        except Exception:
            self._log_config(force_log=True)
            raise

        task_enabled = True
        if self.parent_task and not self.parent_task.ctrl.should_run():
            task_enabled = False

        # load from task_band if exists
        task_band_param = task_params.get_param_value(TASK_BAND_PARAMETER_NAME)
        if task_band_param and task_band_param.value:
            task_band = task_band_param.value

            # we are going to load all task parameters from task_band
            task_params = self.load_task_params_from_task_band(task_band, task_params)

        params = task_params.get_params_signatures(ParameterFilters.SIGNIFICANT_INPUTS)

        # we add override to Object Cache signature
        override_signature = self._get_override_params_signature()
        # task schema id is unique per Class definition.
        # so if we have new implementation - we will not a problem with rerunning it
        full_task_name = "%s@%s(object=%s)" % (
            self.task_name,
            self.task_definition.full_task_family,
            str(id(self.task_definition)),
        )

        # now we don't know the real signature - so we calculate signature based on all known params
        cache_object_signature = build_signature(
            name=full_task_name,
            params=params,
            extra={"task_override": override_signature},
        )
        self._log_build_step(
            "Task task_signature %s" % str(cache_object_signature.signature)
        )

        # If a Task has already been instantiated with the same parameters,
        # the previous instance is returned to reduce number of object instances.
        tic = databand_context.task_instance_cache
        cached_task_object = tic.get_cached_task_obj(cache_object_signature)
        if cached_task_object and not hasattr(cached_task_object, "_dbnd_no_cache"):
            return cached_task_object

        # we want to have task id immediately, so we can initialize outputs/use by user
        # we should switch to SIGNIFICANT_INPUT here
        task_signature_obj = build_signature(
            name=self.task_name,
            params=params,
            extra=self.task_definition.task_signature_extra,
        )

        task_children_scope_params = self._calculate_task_children_scope_params(
            task_params=task_params
        )

        task = task_metaclass._build_task_obj(
            task_definition=self.task_definition,
            task_name=self.task_name,
            task_params=task_params,
            task_signature_obj=task_signature_obj,
            task_config_override=self.task_config_override,
            task_config_layer=self.config.config_layer,
            task_enabled=task_enabled,
            task_sections=self.config_sections,
            task_children_scope_params=task_children_scope_params,
        )
        tic.register_task_obj_cache_instance(
            task, task_obj_cache_signature=cache_object_signature
        )

        task.task_call_source = [
            databand_context.user_code_detector.find_user_side_frame(2)
        ]
        if task.task_call_source and self.parent_task:
            task.task_call_source.extend(self.parent_task.task_call_source)

        # now the task is created - all nested constructors will see it as parent
        with task_context(task, TaskContextPhase.BUILD):
            task._initialize()
            task._validate()

            # it might be that config has been changed even more
            task.task_config_layer = self.config.config_layer

        # only now we know "task_id" so we can register in "publicaly facing cache
        tic.register_task_instance(task)

        return task

    def _update_params_def_target_config(self, param_def):
        """
        calculates parameter.target_config based on extra config
        at `parameter_name__target value
        user might want to change the target_config for specific param using configuration/cli
        """

        # used for target_format update
        # change param_def definition based on config state
        target_def_key = "%s__target" % param_def.name
        target_config = self.config.get_multisection_config_value(
            self.config_sections, key=target_def_key
        )
        if not target_config:
            return param_def

        # the last is from a higher level
        target_config = last(target_config)
        try:
            target_config = parse_target_config(target_config.value)
        except Exception as ex:
            raise param_def.parameter_exception(
                "Calculate target config for %s : target_config='%s'"
                % (target_def_key, target_config.value),
                ex,
            )

        param_def = param_def.modify(target_config=target_config)
        return param_def

    def _build_task_param_values(self):
        """
        This process is composed from those parts:
        1. Adding extra layers of config (based on defaults, task_config and override)
        2. Building the params
        3. Validating the params
        """
        # copy because we about to edit it
        params_to_build = self.task_definition.task_param_defs.copy()
        task_param_values = []

        # let's start to apply layers on current config
        # at the end of the build we will save current layer and will make it "active" config
        # during task execution/initializaiton

        # SETUP: Task may have TaskClass.defaults= {...} - LOWEST priority
        # add these values, so task build or nested tasks/configs builds will see it
        task_defaults = self.task_definition.task_defaults_config_store
        if task_defaults:
            # task_defaults are processed at TaskDefinition,
            # it's ConfigStore with priority = ConfigValuePriority.FALLBACK,
            self.config.set_values(
                config_values=task_defaults, source=self._source_name("task.defaults")
            )

        # SETUP: build and apply any Task class level config
        # User has specified "task_config = ..."
        # Only orchestration tasks has this property
        param_task_config = params_to_build.pop("task_config", None)
        if param_task_config:
            task_param_values.append(
                self._calculate_and_add_layer_of_task_config(param_task_config)
            )

        # SETUP: Apply all "override" values in Task(override={})
        # This is highest priority config, should deprecated in favor of "with config(..)"
        if self.task_config_override:
            self.config.set_values(
                source=self._source_name("ctor[override]"),
                config_values=self.task_config_override,
                priority=ConfigValuePriority.OVERRIDE,
            )

        # SETUP: build and set EnvConfig
        param_task_env = params_to_build.pop("task_env", None)
        if param_task_env:
            # we apply them to config only if there are no values (this is defaults)
            """
            find same parameters in the current task class and EnvConfig
            and take the value of that params from environment
            for example  spark_config in SparkTask  (defined by EnvConfig.spark_config)
            we take values using names only
            """
            value_task_env_config = self._build_parameter_value(param_task_env)
            self.task_env_config = value_task_env_config.value  # type: EnvConfig
            task_param_values.append(value_task_env_config)

        # log with the final config
        self._log_config()

        # calculate configuration per parameter, and calculate parameter value
        for param_def in params_to_build.values():
            param_def = self._update_params_def_target_config(param_def)
            try:
                p_value = self._build_parameter_value(param_def)
                task_param_values.append(p_value)
            except MissingParameterError as ex:
                self.task_errors.append(ex)

        # validation for errors and log warnings
        self.validate_no_extra_config_params(task_param_values)

        if self.task_errors:
            self._raise_task_build_errors()

        if self.build_warnings:
            self._log_task_build_warnings()

        return task_param_values

    def validate_no_extra_config_params(self, task_param_values):
        task_param_values = {param.name: param for param in task_param_values}

        if "validate_no_extra_params" not in task_param_values:
            return

        validate_no_extra_params = task_param_values["validate_no_extra_params"].value
        if validate_no_extra_params == ParamValidation.disabled:
            return

        # Must lower task parameter name to comply to case insensitivity of configuration
        task_param_names = {
            name.lower() for name in self.task_definition.task_param_defs
        }

        for key, value in _iterate_config_items(self.config, self.config_sections):
            if (
                value.source.endswith(
                    format_source_suffix(ParameterScope.children.value)
                )
                or key in ["_type", "_from"]
                or key.endswith("__target")
            ):
                # all those are exceptions we want to ignore
                continue

            if key not in task_param_names:
                # those are keys we need to warn about
                exc = friendly_error.task_build.unknown_parameter_in_config(
                    task_name=self.task_name,
                    param_name=key,
                    source=value.source,
                    task_param_names=task_param_names,
                    config_type=self._get_task_or_config_string(),
                )

                if validate_no_extra_params == ParamValidation.warn:
                    self.build_warnings.append(exc)

                elif validate_no_extra_params == ParamValidation.error:
                    self.task_errors.append(exc)

    def _calculate_task_children_scope_params(self, task_params):
        # type: (Parameters)-> Dict[str, ParameterValue]
        """
        Task can get the value for some params from Parent task (parent of any level)
        If param is defined at parent with scope=ParameterScope.children, it will be added to
        task_children_scope_params and will be used at param calculation

        for example:
          class ChildTask(Task):
              my_p = parameter() # it doesn't have to be scoped
          class ParentTask(Task):
              my_p = parameter(scope=ParameterScope.children)
            def band:
              my_p_task = ChildTask() # no explicit "wiring" of my_p


        """
        if not self.task_cls._conf__scoped_params:
            return {}

        if self.parent_task and self.parent_task.task_children_scope_params:
            # if we have parent - we need to use as a baseline
            task_children_scope_params = (
                self.parent_task.task_children_scope_params.copy()
            )
        else:
            task_children_scope_params = {}

        for p_val in task_params.get_param_values():
            #  we want only parameters of the right scope -- children
            if p_val.parameter.scope == ParameterScope.children:
                task_children_scope_params[p_val.name] = p_val
        return task_children_scope_params

    def _build_and_validate_task_ctor_kwargs(self, task_args, task_kwargs):
        """
        validate is required to handle with parameters with no definition

        this can be caused by function declaration:
        >>> def func(element, *args, k=None, **kwargs): ...

        this cal will create undefined handle for the args
         >>> func(1, problem_1,problem_2 ...)

        this cal will create undefined handle for the kwargs
        >>> func(1, s=value)
        """
        param_names = set(self.task_definition.task_param_defs)
        has_varargs = False
        has_varkwargs = False
        if self.task_cls.task_decorator is not None:
            callable_spec = self.task_cls.task_decorator.get_callable_spec()
            # only in functions we can have args as we know exact "call" signature
            task_args, task_kwargs = args_to_kwargs(
                callable_spec.args, task_args, task_kwargs
            )
            has_varargs = callable_spec.varargs
            has_varkwargs = callable_spec.varkw

        if task_args and not has_varargs:
            # we should not have any args, so we don't know how to assign them
            raise friendly_error.unknown_args_in_task_call(
                self.parent_task,
                self.task_cls,
                self.task_args__ctor,
                call_repr=self._ctor_as_str,
            )

        if not has_varkwargs:
            # there are no kwargs for the callable
            # we need to be sure that all the kwargs are for parameters
            for key in task_kwargs:
                if key not in param_names:
                    raise friendly_error.task_build.unknown_parameter_in_constructor(
                        constructor=self._ctor_as_str,
                        param_name=key,
                        task_parent=self.parent_task,
                    )

        return task_kwargs

    def _calculate_and_add_layer_of_task_config(self, param_task_config):
        """
        calculate any task class level params and updates the config with thier values
        @pipeline(task_config={CoreConfig.tracker: ["console"])
        <or>
        class a(Task):
            task_config = {"core": {"tracker": ["console"]}}
            <or>
            task_config = {CoreConfig.tracker: ["console"]}
        """
        # calculate the value of `Task.task_config` using it's definition
        # (check for all inheritance and find value for `task_config`
        param_task_config_value = self._build_parameter_value(param_task_config)
        if param_task_config_value.value:
            # Support two modes:
            # 1. Task.param_name:333
            # 2. {"section": {"key":"value"}}
            # dict parameter value can't have non string as a key
            param_task_config_value.value = parse_and_build_config_store(
                config_values=param_task_config_value.value,
                source=self._source_name("task_config"),
            )
            # merging `Task.task_config` into current configuration
            # we are adding "ultimate" layer on top of all layers
            self.config.set_values(param_task_config_value.value)

        return param_task_config_value

    def __str__(self):
        if self.task_name == self.task_passport.task_family:
            return "TaskFactory(%s)" % self.task_name
        return "TaskFactory(%s@%s)" % (self.task_name, self.task_family)

    def load_task_params_from_task_band(self, task_band, task_params):
        task_band_value = target(task_band).as_object.read_json()

        new_params = []
        found = []
        source = "task_band.json"
        for p_value in task_params.get_param_values():
            name = p_value.name
            if name not in task_band_value or name == RESULT_PARAM:
                new_params.append(p_value)
                continue

            value = p_value.parameter.calc_init_value(task_band_value[name])
            found.append(name)
            new_parameter_value = ParameterValue(
                parameter=p_value.parameter,
                source=source,
                source_value=value,
                value=value,
            )
            new_params.append(new_parameter_value)

        logger.info(
            "Loading task '{task_family}' from {task_band}:\n"
            "\tfields taken:\t{found}".format(
                task_family=self.task_passport.task_family,
                task_band=task_band,
                found=",".join(found),
            )
        )
        return Parameters("task_band", new_params)

    def _log_build_step(self, msg, force_log=False):
        if self.verbose_build or force_log:
            logger.info("[%s] %s", self.task_name, msg)

    def _log_config(self, force_log=False):
        msg = "config for sections({config_sections}): {config}".format(
            config=pformat_current_config(
                self.config, sections=self.config_sections, as_table=True
            ),
            config_sections=self.config_sections,
        )
        self._log_build_step(msg, force_log=force_log)

    def _raise_task_build_errors(self):
        if len(self.task_errors) == 1:
            raise self.task_errors[0]

        raise friendly_error.failed_to_create_task(
            self._ctor_as_str, nested_exceptions=self.task_errors
        )

    def _get_task_or_config_string(self):
        from dbnd._core.task.config import Config

        if issubclass(self.task_cls, Config):
            return "config"
        else:
            return "task"

    def _source_name(self, name):
        return self.task_definition.task_passport.format_source_name(name)

    def _log_task_build_warnings(self):
        w = "Build warnings for %s '%s': " % (
            self._get_task_or_config_string(),
            self.task_name,
        )

        for warning in self.build_warnings:
            w += "\n\t" + str(warning)
            if hasattr(warning, "did_you_mean") and warning.did_you_mean:
                w += "\n\t\t - " + warning.did_you_mean

        logger.warning(w)


def update_config_section_on_change_only(
    config, config_store, source, on_change_only=False
):
    if on_change_only:
        # we take values using names only
        relevant_config_store = _ConfigStore()
        for section, section_values in six.iteritems(config_store):
            for key, value in six.iteritems(section_values):
                previous_value = config.get_config_value(section, key)
                if (
                    previous_value
                    and on_change_only
                    and previous_value.value == value.value
                ):
                    continue
                relevant_config_store.set_config_value(section, key, value)
        config_store = relevant_config_store

    # we apply set on change only in the for loop, so we can optimize and not run all these code
    if config_store:
        config.set_values(config_values=config_store, source=source)


def _iterate_config_items(config, sections):
    for section_name in sections:
        section = config.config_layer.config.get(section_name)
        if section:
            for key, value in iteritems(section):
                yield key, value


def _get_call_repr(call_name, call_args, call_kwargs):
    params = ""
    if call_args:
        params = ", ".join((safe_string(repr(p), 300)) for p in call_args)
    if call_kwargs:
        if params:
            params += ", "
        params += ", ".join(
            (
                "%s=%s" % (p, safe_string(repr(k), 300))
                for p, k in iteritems(call_kwargs)
            )
        )
    return "{call_name}({params})".format(call_name=call_name, params=params)
